/*

  This file is a part of GRALEEDITOR, a program to help create input data
  for gravitational lens inversion using GRALE.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "editorscene.h"
#include "layeritem.h"
#include "coordinateitem.h"
#include "pointitem.h"
#include "pointmatchitem.h"
#include "triangleitem.h"
#include "graleeditor.h"
#include "fitsitem.h"
#include "imageitem.h"
#include "hullitem.h"
#include "contourleveldialog.h"
#include "contourfinder.h"
#include "rectitem.h"
#include <QGraphicsSceneMouseEvent>
#include <QMessageBox>
#include <QFileDialog>
#include <QPainter>
#include <serut/fileserializer.h>
#include <grale/imagesdata.h>
#include <grale/constants.h>
#include <grale/polygon2d.h>
#include <map>
#include <list>
#include <vector>
#include <cmath>
#include <memory>
#include <iostream>

using namespace grale;
using namespace serut;
using namespace std;

EditorScene::EditorScene()
{
	m_currentLayer = -1;
	m_defaultPointScaleFactor = 0.05;
	m_defaultMatchPointScaleFactor = 0.1;

	m_pCoordinateItem = new CoordinateItem();
	addItem(m_pCoordinateItem);
}

EditorScene::~EditorScene()
{
	clear();
}

void EditorScene::clear()
{
	while (m_layers.size() > 0)
		deleteLayer(0);
}

int EditorScene::addLayer(LayerItem::LayerType layerType)
{
	LayerItem *pLayer = new LayerItem(layerType);

	if (layerType == LayerItem::Points)
		pLayer->setPointScaleFactor(m_defaultPointScaleFactor);
	else
		pLayer->setPointScaleFactor(m_defaultMatchPointScaleFactor);

	m_layers.append(pLayer);

	addItem(pLayer);

	updateZValues();

	setCurrentLayer(m_layers.size()-1);

	return m_layers.size()-1;
}

void EditorScene::setCurrentLayer(int l)
{
	for (int i = 0 ; i < m_layers.size() ; i++)
	{
		LayerItem *pLayer = m_layers.at(i);

		pLayer->disable();
	}

	if (l >= 0 && l < m_layers.size())
	{
		m_currentLayer = l;

		LayerItem *pLayer = m_layers.at(l);

		pLayer->enable();
	}
	else
		m_currentLayer = -1;

}

void EditorScene::updateZValues()
{
	for (int i = 0 ; i < m_layers.size() ; i++)
	{
		LayerItem *pLayer = m_layers.at(i);

		pLayer->setZValue(i);
	}
}

void EditorScene::decreaseLayerIndex(int l)
{
	if (l < 1 || l >= m_layers.size())
		return;

	m_layers.swap(l-1, l);
	updateZValues();
}

void EditorScene::increaseLayerIndex(int l)
{
	if (l < 0 || l >= m_layers.size()-1)
		return;

	m_layers.swap(l, l+1);
	updateZValues();
}

void EditorScene::mousePressEvent(QGraphicsSceneMouseEvent *mouseEvent)
{
	if (m_currentLayer < 0 || m_currentLayer >= m_layers.size())
		return;

	LayerItem *pLayer = m_layers.at(m_currentLayer);
	GraleEditor::Tool currentTool = GraleEditor::getInstance()->getCurrentTool();
	QPointF point = mouseEvent->scenePos();
	bool useDefault = true;

	if (currentTool == GraleEditor::AddPointTool)
	{
		if (pLayer->getLayerType() == LayerItem::Points && pLayer->isVisible() && mouseEvent->button() == Qt::LeftButton)
			pLayer->addPoint(point);
	}
	else if (currentTool == GraleEditor::HullTool)
	{
		if (pLayer->getLayerType() == LayerItem::Points && pLayer->isVisible() && mouseEvent->button() == Qt::LeftButton)
		{
			pLayer->startHull(point);
			useDefault = false;
			mouseEvent->accept();
		}
	}

	if (useDefault)
		QGraphicsScene::mousePressEvent(mouseEvent);
}

void EditorScene::mouseReleaseEvent(QGraphicsSceneMouseEvent *mouseEvent)
{
	if (m_currentLayer < 0 || m_currentLayer >= m_layers.size())
		return;

	bool useDefault = true;
	LayerItem *pLayer = m_layers.at(m_currentLayer);
	GraleEditor::Tool currentTool = GraleEditor::getInstance()->getCurrentTool();

	if (currentTool == GraleEditor::HullTool)
	{
		if (pLayer->getLayerType() == LayerItem::Points)
		{
			pLayer->endHull();
			useDefault = false;
			mouseEvent->accept();
		}
	}

	if (useDefault)
		QGraphicsScene::mouseReleaseEvent(mouseEvent);
}

void EditorScene::mouseMoveEvent(QGraphicsSceneMouseEvent *mouseEvent)
{
	if (m_currentLayer < 0 || m_currentLayer >= m_layers.size())
		return;

	bool useDefault = true;
	LayerItem *pLayer = m_layers.at(m_currentLayer);
	GraleEditor::Tool currentTool = GraleEditor::getInstance()->getCurrentTool();
	QPointF point = mouseEvent->scenePos();

	if (currentTool == GraleEditor::HullTool)
	{
		if (pLayer->getLayerType() == LayerItem::Points && pLayer->isDrawingHull())
		{
			pLayer->addHullPoint(point);
			useDefault = false;
			mouseEvent->accept();
		}
	}

	if (useDefault)
		QGraphicsScene::mouseMoveEvent(mouseEvent);
}

void EditorScene::objectFinder(QPointF point, LayerItem *pLayer)
{
	// TODO: make these things configurable
	int pixelSize = 256;
	double viewSize = 4.0; // in arcsec

	QPixmap pixmap(pixelSize, pixelSize);

	pixmap.fill(Qt::black);
	QPainter painter(&pixmap);

	setCaching(false);

	int numLayers = getNumberOfLayers();
	vector<bool> visibleFlags(numLayers);

	for (int i = 0 ; i < numLayers ; i++)
	{
		LayerItem *pLayer = getLayerItem(i);
		visibleFlags[i] = pLayer->isVisible();

		if (pLayer->getLayerType() != LayerItem::Image)
			pLayer->setVisible(false);
	}

	QRectF region(point.x()-viewSize/2.0, point.y()-viewSize/2.0, viewSize, viewSize);

	painter.setRenderHint(QPainter::Antialiasing);
	render(&painter, QRectF(0, 0, pixelSize, pixelSize), region);
	painter.end();

	setCaching(true);
	for (int i = 0 ; i < numLayers ; i++)
	{
		LayerItem *pLayer = getLayerItem(i);
		pLayer->setVisible(visibleFlags[i]);
	}

	findObjectAtCenter(region, pixmap, pLayer);
}

void EditorScene::mouseDoubleClickEvent(QGraphicsSceneMouseEvent *mouseEvent)
{
	if (m_currentLayer < 0 || m_currentLayer >= m_layers.size())
		return;

	LayerItem *pLayer = m_layers.at(m_currentLayer);
	if (pLayer->getLayerType() == LayerItem::Points && pLayer->isVisible() && mouseEvent->button() == Qt::LeftButton)
	{
		mouseEvent->accept();
		QPointF point = mouseEvent->scenePos();
		//cout << "Double click: " << point.x() << "," << point.y() << endl;

		objectFinder(point, pLayer);
	}
	else
		QGraphicsScene::mouseDoubleClickEvent(mouseEvent);
}

void EditorScene::deleteLayer(int l)
{
	if (l < 0 || l >= m_layers.size())
		return;

	LayerItem *pLayer = m_layers.at(l);
	m_layers.removeAt(l);

	// Need to delete the triangles first so that they don't refer to invalid points
	pLayer->deleteTriangles();
	delete pLayer;

	if (m_currentLayer >= m_layers.size())
		m_currentLayer = m_layers.size()-1;
}

bool EditorScene::importImagesData(const QString &fname, QList<LayerItem *> &newLayers)
{
	ImagesData imgDat;

	if (!imgDat.load(fname.toStdString()))
	{
		setErrorString(imgDat.getErrorString());
		return false;
	}

	newLayers.clear();

	int numImages = imgDat.getNumberOfImages();
	std::vector<std::vector<PointItem *> > pointItems(numImages);

	for (int i = 0 ; i < numImages ; i++)
	{
		int newLayer = addLayer(LayerItem::Points);
		LayerItem *pLayer = getLayerItem(newLayer);
		int numPoints = imgDat.getNumberOfImagePoints(i);

		pointItems[i].resize(numPoints);
		newLayers.append(pLayer);

		for (int p = 0 ; p < numPoints ; p++)
		{
			Vector2D<double> pos = imgDat.getImagePointPosition(i, p)/ANGLE_ARCSEC;
			PointItem *pPoint = new PointItem(pLayer);
			
			pPoint->setPos(pos.getX(), pos.getY());
			pPoint->setScaleFactor(pLayer->getPointScaleFactor());

			pointItems[i][p] = pPoint;

			if (imgDat.hasIntensities())
				pPoint->setIntensity(imgDat.getImagePointIntensity(i, p));
		}

		if (imgDat.hasTriangulation())
		{
			std::list<TriangleIndices> triangles;
			std::list<TriangleIndices>::const_iterator it;

			if (imgDat.getTriangles(i, triangles))
			{
				for (it = triangles.begin() ; it != triangles.end() ; it++)
				{
					TriangleIndices indices = *it;

					TriangleItem *pTriangle = new TriangleItem(pointItems[i][indices.getIndex(0)],
							                           pointItems[i][indices.getIndex(1)],
										   pointItems[i][indices.getIndex(2)],
										   pLayer);
				}
			}
		}
	}

	for (int g = 0 ; g < imgDat.getNumberOfGroups() ; g++)
	{
		int numGroupPoints = imgDat.getNumberOfGroupPoints(g);

		for (int gp = 0 ; gp < numGroupPoints ; gp++)
		{
			int img, point;

			imgDat.getGroupPointIndices(g, gp, &img, &point);

			pointItems[img][point]->setGroupNumber(g);
		}
	}

	if (imgDat.hasTimeDelays())
	{
		int numDelays = imgDat.getNumberOfTimeDelays();

		for (int i = 0 ; i < numDelays ; i++)
		{
			int img, point;
			double delay;

			imgDat.getTimeDelay(i, &img, &point, &delay);

			pointItems[img][point]->setTimeDelay(delay);
		}
	}

	return true;
}

bool EditorScene::exportImagesData(const QString &fname, bool triangles, bool intensities, bool groups, bool timeDelays)
{
	int numLayers = getNumberOfLayers();
	QList<LayerItem *> pointLayers;

	for (int i = 0 ; i < numLayers ; i++)
	{
		LayerItem *pLayer = getLayerItem(i);

		if (pLayer->isVisible() && pLayer->getLayerType() == LayerItem::Points)
			pointLayers.append(pLayer);
	}

	if (pointLayers.size() == 0)
	{
		setErrorString("No visible point layers to save");
		return false;
	}

	ImagesData imgDat;

	if (!imgDat.create(pointLayers.size(), intensities, false))
	{
		setErrorString(imgDat.getErrorString());
		return false;
	}

	std::map<PointItem *, std::pair<int, int> > pointMap;
	std::map<int, std::list<PointItem *> > groupPoints;

	for (int i = 0 ; i < pointLayers.size() ; i++)
	{
		LayerItem *pLayer = pointLayers.at(i);
		QList<PointItem *> pointList;
		
		pointList = pLayer->getAllPoints();

		for (int p = 0 ; p < pointList.size() ; p++)
		{
			PointItem *pPoint = pointList.at(p);
			QPointF pos = pPoint->pos();

			Vector2D<double> angularPos(pos.x()*ANGLE_ARCSEC, pos.y()*ANGLE_ARCSEC);
			double intens = pPoint->getIntensity();

			int pointIndex;
			
			if (intensities)
				pointIndex = imgDat.addPoint(i, angularPos, intens);
			else
				pointIndex = imgDat.addPoint(i, angularPos);

			if (pointIndex < 0)
			{
				setErrorString(imgDat.getErrorString());
				return false;
			}

			pointMap[pPoint] = std::pair<int,int>(i, pointIndex);

			if (groups)
			{
				if (pPoint->hasGroupNumber())
				{
					int groupNumber = pPoint->getGroupNumber();

					groupPoints[groupNumber].push_back(pPoint);
				}
			}
			if (timeDelays)
			{
				if (pPoint->hasTimeDelay())
				{
					if (!imgDat.addTimeDelayInfo(i, pointIndex, pPoint->getTimeDelay()))
					{
						setErrorString(imgDat.getErrorString());
						return false;
					}
				}
			}
		}

		if (triangles)
		{
			QList<TriangleItem *> triangleList;

			triangleList = pLayer->getAllTriangles();

			for (int t = 0 ; t < triangleList.size() ; t++)
			{
				TriangleItem *pTriangle = triangleList.at(t);
				int indices[3];

				for (int p = 0 ; p < 3 ; p++)
				{
					PointItem *pPoint = pTriangle->getPoint(p);

					indices[p] = pointMap[pPoint].second;
				}

				if (!imgDat.addTriangle(i, indices[0], indices[1], indices[2]))
				{
					setErrorString(imgDat.getErrorString());
					return false;
				}
			}
		}
	}

	if (groups)
	{
		std::map<int, std::list<PointItem *> >::iterator it;

		for (it = groupPoints.begin() ; it != groupPoints.end() ; it++)
		{
			std::list<PointItem *> &groupPointList = (*it).second;

			int groupIndex = imgDat.addGroup();

			if (groupIndex < 0)
			{
				setErrorString(imgDat.getErrorString());
				return false;
			}

			std::list<PointItem *>::const_iterator it2;

			for (it2 = groupPointList.begin() ; it2 != groupPointList.end() ; it2++)
			{
				PointItem *pPoint = *it2;

				std::pair<int, int> pointIndices = pointMap[pPoint];

				if (!imgDat.addGroupPoint(groupIndex, pointIndices.first, pointIndices.second))
				{
					setErrorString(imgDat.getErrorString());
					return false;
				}
			}
		}
	}

	if (!imgDat.save(fname.toStdString()))
	{
		setErrorString(imgDat.getErrorString());
		return false;
	}

	return true;
}

void EditorScene::enableMoveAndSelect(bool move)
{
	disableMoveAndSelect();

	if (m_currentLayer < 0)
		return;

	LayerItem *pLayer = m_layers.at(m_currentLayer);

	if (pLayer->getLayerType() == LayerItem::Points)
	{
		QList<PointItem *> pointList;
	
		pointList = pLayer->getAllPoints();
		for (int i = 0 ; i < pointList.size() ; i++)
		{
			PointItem *pPoint = pointList.at(i);

			if (move)
				pPoint->setFlag(QGraphicsItem::ItemIsMovable);
			pPoint->setFlag(QGraphicsItem::ItemIsSelectable);
		}

		QList<TriangleItem *> triangleList;

		triangleList = pLayer->getAllTriangles();
		for (int i = 0 ; i < triangleList.size() ; i++)
		{
			TriangleItem *pTriangle = triangleList.at(i);

			pTriangle->setFlag(QGraphicsItem::ItemIsSelectable);
		}
	}
	else if (pLayer->getLayerType() == LayerItem::Image)
	{
		QList<PointMatchItem *> pointList;
		
		pointList = pLayer->getAllMatchPoints();
		for (int i = 0 ; i < pointList.size() ; i++)
		{
			PointMatchItem *pPoint = pointList.at(i);

			if (move)
				pPoint->setFlag(QGraphicsItem::ItemIsMovable);
			pPoint->setFlag(QGraphicsItem::ItemIsSelectable);
		}

		QList<QGraphicsItem *> children = pLayer->childItems();
		for (int i = 0 ; i < children.size() ; i++)
		{
			QGraphicsItem *pChild = children.at(i);

			pChild->setFlag(QGraphicsItem::ItemIsSelectable);
		}
	}
}

void EditorScene::disableMoveAndSelect()
{
	int numLayers = getNumberOfLayers();

	for (int i = 0 ; i < numLayers ; i++)
	{
		LayerItem *pLayer = getLayerItem(i);

		if (pLayer->getLayerType() == LayerItem::Points)
		{
			QList<PointItem *> pointList;
		
			pointList = pLayer->getAllPoints();
			for (int i = 0 ; i < pointList.size() ; i++)
			{
				PointItem *pPoint = pointList.at(i);

				pPoint->setFlag(QGraphicsItem::ItemIsMovable, false);
				pPoint->setFlag(QGraphicsItem::ItemIsSelectable, false);
			}

			QList<TriangleItem *> triangleList;

			triangleList = pLayer->getAllTriangles();
			for (int i = 0 ; i < triangleList.size() ; i++)
			{
				TriangleItem *pTriangle = triangleList.at(i);

				pTriangle->setFlag(QGraphicsItem::ItemIsSelectable, false);
			}
		}
		else if (pLayer->getLayerType() == LayerItem::Image)
		{
			QList<PointMatchItem *> pointList;
		
			pointList = pLayer->getAllMatchPoints();
			for (int i = 0 ; i < pointList.size() ; i++)
			{
				PointMatchItem *pPoint = pointList.at(i);

				pPoint->setFlag(QGraphicsItem::ItemIsMovable, false);
				pPoint->setFlag(QGraphicsItem::ItemIsSelectable, false);
			}

			QList<QGraphicsItem *> children = pLayer->childItems();
			for (int i = 0 ; i < children.size() ; i++)
			{
				QGraphicsItem *pChild = children.at(i);

				pChild->setFlag(QGraphicsItem::ItemIsSelectable, false);
			}
		}
	}
}

bool EditorScene::load(const QString &fname)
{
	FileSerializer fSer;

	if (!fSer.open(fname.toStdString(), FileSerializer::ReadOnly))
	{
		setErrorString(fSer.getErrorString());
		return false;
	}
	return load(fSer);
}

bool EditorScene::save(const QString &fname) const
{
	FileSerializer fSer;

	if (!fSer.open(fname.toStdString(), FileSerializer::WriteOnly))
	{
		setErrorString(fSer.getErrorString());
		return false;
	}
	return save(fSer);
}

#define ID 0x44455247
#define LAYER_POINTS 1
#define LAYER_IMAGE 2
#define IMAGELAYER_EMPTY 0
#define IMAGELAYER_FITS 1
#define IMAGELAYER_RGB 2

bool EditorScene::load(SerializationInterface &si)
{
	int32_t id, numLayers, currentLayer;

	clear();

	if ( !(si.readInt32(&id) && 
	       si.readInt32(&numLayers) &&
	       si.readInt32(&currentLayer) ) )
	{
		setErrorString(si.getErrorString());
		return false;
	}
	if (id != ID)
	{
		setErrorString("Read invalid file identifier");
		return false;
	}

	for (int32_t i = 0 ; i < numLayers ; i++)
	{
		int32_t layerID;
		double pointScaleFactor;
		int32_t nextGroupNumber;
		int32_t visible;

		if (! (si.readInt32(&layerID) &&
		       si.readDouble(&pointScaleFactor) &&
		       si.readInt32(&nextGroupNumber) &&
		       si.readInt32(&visible) ) )
		{
			setErrorString(si.getErrorString());
			return false;
		}

		LayerItem::LayerType layerType;

		if (layerID == LAYER_POINTS)
			layerType = LayerItem::Points;
		else if (layerID == LAYER_IMAGE)
			layerType = LayerItem::Image;
		else
		{
			setErrorString("Read invalid layer type");
			return false;
		}

		int idx = addLayer(layerType);
		LayerItem *pLayer = getLayerItem(idx);

		if (layerType == LayerItem::Points)
		{
			int32_t numPoints;
			int32_t numTriangles;

			QList<PointItem *> points;

			if (! (si.readInt32(&numPoints) &&
			       si.readInt32(&numTriangles) ) )
			{
				setErrorString(si.getErrorString());
				return false;
			}

			for (int32_t i = 0 ; i < numPoints ; i++)
			{
				double xy[2], intensity, timeDelay;
				int32_t timeFlag, groupNumber;

				if (! (si.readDoubles(xy, 2) &&
				       si.readInt32(&timeFlag) &&
				       si.readInt32(&groupNumber) &&
				       si.readDouble(&intensity) &&
				       si.readDouble(&timeDelay) ) )
				{
					setErrorString(si.getErrorString());
					return false;
				}

				PointItem *pPoint = new PointItem(pLayer);
				pPoint->setPos(xy[0], xy[1]);
				pPoint->setGroupNumber(groupNumber);
				pPoint->setIntensity(intensity);
				if (timeFlag)
					pPoint->setTimeDelay(timeDelay);
				points.append(pPoint);
			}
			for (int32_t i = 0 ; i < numTriangles ; i++)
			{
				int32_t indices[3];

				if (!si.readInt32s(indices, 3))
				{
					setErrorString(si.getErrorString());
					return false;
				}
				TriangleItem *pTriangle = new TriangleItem(points.at(indices[0]),
						                           points.at(indices[1]),
									   points.at(indices[2]),
									   pLayer);
			}
		}
		else // image
		{
			int32_t imageLayerType;

			if (!si.readInt32(&imageLayerType))
			{
				setErrorString(si.getErrorString());
				return false;
			}

			QGraphicsItem *pItem = 0;

			if (imageLayerType == IMAGELAYER_EMPTY)
			{
				// nothing to do
			}
			else if (imageLayerType == IMAGELAYER_FITS)
			{
				std::string fileName;
				Vector2D<double> center;
				double	min, max;

				if (! (si.readString(fileName) &&
				       si.readDoubles(center.getComponents(), 2) &&
				       si.readDouble(&min) &&
				       si.readDouble(&max) ) )
				{
					setErrorString(si.getErrorString());
					return false;
				}

				FITSItem *pFITSItem = new FITSItem(pLayer);	

				bool done = false;
				bool success = false;

				while (!done)
				{
					if (pFITSItem->load(QString(fileName.c_str())))
					{
						done = true;
						success = true;
					}
					else
					{
						QMessageBox::warning(GraleEditor::getInstance(), "Warning", QString("Couldn't open FITS file ") + QString(fileName.c_str()));
						QString fname = QFileDialog::getOpenFileName(GraleEditor::getInstance(), "Select FITS file to open", "", "FITS files (*.fit *.fits)");

						if (fname.isNull())
						{
							done = true;
							success = false;
						}
						else
							fileName = fname.toStdString();
					}
				}

				if (!success)
					delete pFITSItem;
				else
				{
					pFITSItem->setMinMax(min, max);
					pFITSItem->setCenterPoint(center);	
					pItem = pFITSItem;
				}		
			}
			else if (imageLayerType == IMAGELAYER_RGB)
			{
				std::string fileName;
				double m[6];

				if (! (si.readString(fileName) &&
				       si.readDoubles(m, 6) ) )
				{
					setErrorString(si.getErrorString());
					return false;
				}

				ImageItem *pImageItem = new ImageItem(pLayer);

				bool done = false;
				bool success = false;

				while (!done)
				{
					if (pImageItem->load(QString(fileName.c_str())))
					{
						done = true;
						success = true;
					}
					else
					{
						QMessageBox::warning(GraleEditor::getInstance(), "Warning", QString("Couldn't open image file ") + QString(fileName.c_str()));
						QString fname = QFileDialog::getOpenFileName(GraleEditor::getInstance(), "Select image file to open", "", "FITS files (*.png *.jpg)");

						if (fname.isNull())
						{
							done = true;
							success = false;
						}
						else
							fileName = fname.toStdString();
					}
				}

				if (!success)
					delete pImageItem;
				else
				{
					pImageItem->setPos(0, 0);
					pImageItem->setTransform(QTransform(m[0], m[1], m[2], m[3], m[4], m[5]), false);
					pItem = pImageItem;
				}
			}
			else
			{
				setErrorString("Read invalid image layer type");
				return false;
			}

			int32_t numPoints;

			if (!si.readInt32(&numPoints))
			{
				setErrorString(si.getErrorString());
				return false;
			}

			for (int32_t i = 0 ; i < numPoints ; i++)
			{
				double xy[2];
				int32_t groupID;

				if (! (si.readDoubles(xy, 2) &&
				       si.readInt32(&groupID) ) )
				{
					setErrorString(si.getErrorString());
					return false;
				}

				if (pItem)
				{
					PointMatchItem *pPoint = new PointMatchItem(pItem);

					pPoint->setPos(xy[0], xy[1]);
					pPoint->setGroupNumber(groupID);
				}
			}
		}

		pLayer->setPointScaleFactor(pointScaleFactor);
		pLayer->setGroupNumber(nextGroupNumber);

		if (visible)
			pLayer->setVisible(true);
		else
			pLayer->setVisible(false);
	}

	setCurrentLayer(currentLayer);
	return true;
}

bool EditorScene::save(SerializationInterface &si) const
{
	if (m_layers.size() == 0)
	{
		setErrorString("No layers to save");
		return false;
	}

	if (! (si.writeInt32(ID) &&
	       si.writeInt32(m_layers.size()) &&
	       si.writeInt32(m_currentLayer) ) )
	{
		setErrorString(si.getErrorString());
		return false;
	}

	for (int i = 0 ; i < m_layers.size() ; i++)
	{
		LayerItem *pLayer = m_layers.at(i);
		int layerID = 0;
		int visible = 0;

		if (pLayer->getLayerType() == LayerItem::Points)
			layerID = LAYER_POINTS;
		else
			layerID = LAYER_IMAGE;
		
		if (pLayer->isVisible())
			visible = 1;

		if (! (si.writeInt32(layerID) &&
		       si.writeDouble(pLayer->getPointScaleFactor()) &&
		       si.writeInt32(pLayer->getNextGroupNumber()) &&
		       si.writeInt32(visible) ) )
		{
			setErrorString(si.getErrorString());
			return false;
		}

		if (pLayer->getLayerType() == LayerItem::Points)
		{
			QList<PointItem *> points = pLayer->getAllPoints();
			QList<TriangleItem *> triangles = pLayer->getAllTriangles();
			std::map<PointItem *, int> pointIndices;

			if (! (si.writeInt32(points.size()) &&
			       si.writeInt32(triangles.size()) ))
			{
				setErrorString(si.getErrorString());
				return false;
			}

			for (int i = 0 ; i < points.size() ; i++)
			{
				PointItem *pPoint = points.at(i);
				QPointF pos = pPoint->pos();
				double xy[2] = { pos.x(), pos.y() };
				int timeFlag = 0;

				pointIndices[pPoint] = i;

				if (pPoint->hasTimeDelay())
					timeFlag = 1;
				if (! (si.writeDoubles(xy, 2) &&
				       si.writeInt32(timeFlag) &&
				       si.writeInt32(pPoint->getGroupNumber()) &&
				       si.writeDouble(pPoint->getIntensity()) &&
				       si.writeDouble(pPoint->getTimeDelay()) ) )
				{
					setErrorString(si.getErrorString());
					return false;
				}
			}

			for (int i = 0 ; i < triangles.size() ; i++)
			{
				TriangleItem *pTriangle = triangles.at(i);
				int32_t indices[3];

				for (int j = 0 ; j < 3 ; j++)
					indices[j] = pointIndices[pTriangle->getPoint(j)];

				if (!si.writeInt32s(indices, 3))
				{
					setErrorString(si.getErrorString());
					return false;
				}
			}
		}
		else // Image layer
		{
			QList<QGraphicsItem *> children = pLayer->childItems();

			if (children.size() == 0)
			{
				if (! (si.writeInt32(IMAGELAYER_EMPTY) &&
				       si.writeInt32(0) ) ) // no point match items
				{
					setErrorString(si.getErrorString());
					return false;
				}
			}
			else 
			{
				QGraphicsItem *pChild = children.at(0);
				FITSItem *pFITSItem = 0;
				ImageItem *pImageItem = 0;

				if ((pFITSItem = dynamic_cast<FITSItem *>(pChild)) != 0)
				{
					Vector2D<double> center = pFITSItem->getCenterPoint();
					std::string fileName = pFITSItem->getFileName().toStdString();
					double min, max;

					pFITSItem->getMinMax(min, max);

					if (! (si.writeInt32(IMAGELAYER_FITS) &&
					       si.writeString(fileName) &&
					       si.writeDoubles(center.getComponents(), 2) &&
					       si.writeDouble(min) &&
					       si.writeDouble(max) ) )
					{
						setErrorString(si.getErrorString());
						return false;
					}
				}
				else if ((pImageItem = dynamic_cast<ImageItem *>(pChild)) != 0)
				{
					std::string fileName = pImageItem->getFileName().toStdString();
					QTransform t = pImageItem->transform();
					double m[6] = { t.m11(), t.m12(), t.m21(), t.m22(), t.dx(), t.dy() };

					if (! (si.writeInt32(IMAGELAYER_RGB) &&
					       si.writeString(fileName) &&
					       si.writeDoubles(m, 6) ) )
					{
						setErrorString(si.getErrorString());
						return false;
					}
				}
				else
				{
					setErrorString("Internal error: unknown image item in layer");
					return false;
				}

				QList<PointMatchItem *> points = pLayer->getAllMatchPoints();
				
				if (!si.writeInt32(points.size()))
				{
					setErrorString(si.getErrorString());
					return false;
				}

				for (int i = 0 ; i < points.size() ; i++)
				{
					PointMatchItem *pPoint = points.at(i);
					QPointF pos = pPoint->pos();
					double xy[2] = { pos.x(), pos.y() };
					int groupID = pPoint->getGroupNumber();

					if (! (si.writeDoubles(xy, 2) &&
					       si.writeInt32(groupID) ) )
					{
						setErrorString(si.getErrorString());
						return false;
					}
				}
			}
		}
	}

	return true;
}

void EditorScene::toggleCoordinateItem()
{
	bool visible = m_pCoordinateItem->isVisible();

	m_pCoordinateItem->setVisible(!visible);
}

void EditorScene::setCaching(bool flag)
{
	int numLayers = getNumberOfLayers();

	for (int i = 0 ; i < numLayers ; i++)
	{
		LayerItem *pLayer = m_layers.at(i);

		if (pLayer->getLayerType() == LayerItem::Image)
		{
			QList<QGraphicsItem *> children = pLayer->childItems();

			for (int i = 0 ; i < children.size() ; i++)
			{
				QGraphicsItem *pChild = children.at(i); // is a FITS item or image item

				if (flag)
					pChild->setCacheMode(QGraphicsItem::DeviceCoordinateCache);
				else
					pChild->setCacheMode(QGraphicsItem::NoCache);
			}
		}
	}
}

void EditorScene::setDefaultPointScaleFactor(double v)
{
	m_defaultPointScaleFactor = v;
}

void EditorScene::setDefaultMatchPointScaleFactor(double v)
{
	m_defaultMatchPointScaleFactor = v;
}

void EditorScene::setAllPointSizes(double s)
{
	int numLayers = getNumberOfLayers();

	for (int i = 0 ; i < numLayers ; i++)
	{
		LayerItem *pLayer = getLayerItem(i);

		if (pLayer->getLayerType() == LayerItem::Points)
			pLayer->setPointScaleFactor(s);
	}
}

void EditorScene::setAllMatchPointSizes(double s)
{
	int numLayers = getNumberOfLayers();

	for (int i = 0 ; i < numLayers ; i++)
	{
		LayerItem *pLayer = getLayerItem(i);

		if (pLayer->getLayerType() == LayerItem::Image)
			pLayer->setPointScaleFactor(s);
	}
}

vector<double> blur(const vector<double> &values, int w, int h, int kernelWidth)
{
	const int kernelSize = 2*kernelWidth+1;
	const int kernelCenter = kernelWidth;
	vector<double> kernel(kernelSize*kernelSize);

	double norm = 0;
#if 0
	// Gaussian like kernel
	for (int y = 0 ; y < kernelSize ; y++)
	{
		double dy = y-kernelCenter;
		for (int x = 0 ; x < kernelSize ; x++)
		{
			double dx = x-kernelCenter;
			double val = exp(-(dx*dx+dy*dy)/(2.0*(kernelWidth*3)));
			norm += val;
			kernel[x+y*kernelSize] = val;
		}
	}
#else
	// Uniform kernel
	for (int y = 0 ; y < kernelSize ; y++)
	{
		for (int x = 0 ; x < kernelSize ; x++)
		{
			double val = 1.0;
			norm += val;
			kernel[x+y*kernelSize] = val;
		}
	}
#endif

	for (auto &x : kernel)
		x /= norm;

	// TODO: check negative!
	const int resultW = w - kernelWidth*2;
	const int resultH = h - kernelWidth*2;
	vector<double> convolution(resultW*resultH);
	for (int y = 0 ; y < resultH ; y++)
	{
		for (int x = 0 ; x < resultW ; x++)
		{
			double sum = 0;
			for (int ky = 0 ; ky < kernelSize ; ky++)
			{
				for (int kx = 0 ; kx < kernelSize ; kx++)
				{
					sum += kernel[kx+ky*kernelSize]*values[(x+kx)+(y+ky)*w];
				}
			}
			convolution[x+y*resultW] = sum;
		}
	}

	return convolution;
}

void EditorScene::findObjectAtCenter(QRectF region, const QPixmap &pixmap, LayerItem *pLayer)
{
	QImage img = pixmap.toImage().mirrored(true, true);
	//img.save("testimage.png"); // TODO: disable this

	int w = img.width();
	int h = img.height();
	vector<double> values(w*h);

	for (int y = 0 ; y < h ; y++)
	{
		for (int x = 0 ; x < w ; x++)
			values[x+y*w] = qGray(img.pixel(x, y));
	}

	int kWidth = 3; // TODO: make this configurable -> in the dialog?
	values = blur(values, w, h, kWidth);
	w -= kWidth*2;
	h -= kWidth*2;

	double x0 = std::min(region.left(), region.right());
	double x1 = std::max(region.left(), region.right());
	double y0 = std::min(region.bottom(), region.top());
	double y1 = std::max(region.bottom(), region.top());
	double pixelWidth = std::abs(x1-x0)/w;
	double pixelHeight = std::abs(y1-y0)/h;

	//cout << "x0 = " << x0 << " x1 = " << x1 << " y0 = " << y0 << " y1 = " << y1 << endl;
	x0 += pixelWidth/2.0;
	x1 -= pixelWidth/2.0;
	y0 += pixelHeight/2.0;
	y1 -= pixelHeight/2.0;
	//cout << "x0 = " << x0 << " x1 = " << x1 << " y0 = " << y0 << " y1 = " << y1 << endl;

	Vector2Dd bottomLeft, topRight;

	bottomLeft = Vector2Dd(x1, y1);
	topRight = Vector2Dd(x0, y0);
	ContourFinder cf(values, bottomLeft, topRight, w, h);
	Vector2Dd center((x0+x1)/2.0, (y0+y1)/2.0);
	vector<vector<Vector2Dd>> contours(256);
	vector<double> averageDistances(contours.size());

	for (int i = 0 ; i < contours.size() ; i++)
	{
		if (!cf.findContours(0.5 + i))
		{
			cerr << "Error finding contours: " << cf.getErrorString() << endl;
			return;
		}

		vector<vector<Vector2Dd>> ctrList = cf.getContours();
		double bestAvgDist = numeric_limits<double>::max();
		vector<Vector2Dd> closestContour;

		for (const vector<Vector2Dd> &ctr : ctrList)
		{
			// Check that the central point is inside the contour
			Polygon2D<double> poly;

			poly.init(&ctr[0], ctr.size(), true); // TODO: make convex hull configurable?
			if (!poly.isInside(center))
				continue;

			auto pPolyPoints = poly.getPoints();
			vector<Vector2Dd> ctrPoly;
			for (int i = 0 ; i < poly.getNumberOfPoints() ; i++)
				ctrPoly.push_back(pPolyPoints[i]);

			double avgDist = 0;
			for (Vector2Dd pos : ctrPoly)
			{
				Vector2Dd diff = pos;
				diff -= center;

				avgDist += diff.getLength();
				//cout << pos.getX() << " " << pos.getY() << endl;
			}
			avgDist /= ctrPoly.size();

			//cout << endl;

			if (avgDist < bestAvgDist)
			{
				bestAvgDist = avgDist;
				closestContour = ctrPoly;
			}
		}

		contours[i] = closestContour;
		averageDistances[i] = bestAvgDist;
		//cout << endl;
	}

	// TODO: use this to estimate a start contour level
	//for (int i = 0 ; i < contours.size() ; i++)
	//{
	//	if (averageDistances[i] < numeric_limits<double>::max())
	//		cout << averageDistances[i] << " " << (0.5+i) << endl;
	//}
	//cout << endl << endl;

	QVector<QVector<QPointF>> qContours;
	for (int i = 0 ; i < contours.size() ; i++)
	{
		const vector<Vector2Dd> &ctr = contours[i];
		QVector<QPointF> qCtr;
		
		for (Vector2Dd pos : ctr)
			qCtr.push_back(QPointF(pos.getX(), pos.getY()));		

		qContours.push_back(qCtr);
	}

	unique_ptr<HullItem> pHull(new HullItem(pLayer));
	unique_ptr<RectItem> pRectItem(new RectItem(region, pLayer));
	QVector<QPointF> hullPoints;

	ContourLevelDialog dlg(GraleEditor::getInstance(), qContours, pHull.get());

	if (dlg.foundContour())
	{
		if (dlg.exec() == QDialog::Accepted)
			hullPoints = pHull->getHullPoints();

		if (hullPoints.size() > 0)
		{
			pLayer->startHull(hullPoints[0]);
			for (int i = 1 ; i < hullPoints.size() ; i++)
				pLayer->addHullPoint(hullPoints[i]);
			pLayer->endHull();
		}
	}
	else
	{
		QMessageBox::warning(GraleEditor::getInstance(), "Error", "No useful contour detected in region");
	}
}

void EditorScene::setCoordinateItem(bool show)
{
	m_pCoordinateItem->setVisible(show);
}

bool EditorScene::isCoordinateItemShown() const
{
	return m_pCoordinateItem->isVisible();
}

