/*

  This file is a part of GRALEEDITOR, a program to help create input data
  for gravitational lens inversion using GRALE.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "fitsitem.h"
#include "graleeditor.h"
#include "pointmatchitem.h"
#include "layeritem.h"
#include "matchpointdialog.h"
#include <grale/fitsimage.h>
#include <grale/coordinateinterface.h>
#include <grale/realtype.h>
#include <grale/constants.h>
#include <QImage>
#include <QPixmap>
#include <QPainter>
#include <QGraphicsSceneMouseEvent>
#include <iostream>

using namespace grale;
using namespace std;

FITSItem::FITSItem(QGraphicsItem *pParent) : QGraphicsItem(pParent)
{
	m_pFITSImage = 0;
	m_pPixmap = 0;
	m_min = 0;
	m_max = 1;
	m_pCoordinates = 0;

	setCacheMode(QGraphicsItem::DeviceCoordinateCache);
//	setFlag(ItemIsSelectable);
}

FITSItem::~FITSItem()
{
	if (m_pFITSImage)
		delete m_pFITSImage;
	if (m_pPixmap)
		delete m_pPixmap;
}

bool FITSItem::load(const QString &fname)
{
	FITSImage *pFITS = new FITSImage();

	if (pFITS->loadWCSCoordinates(fname.toStdString()))
	{
//		std::cerr << pFITS->getErrorString() << std::endl;
//		delete pFITS;
//		return false;

		clear();

		cout << "Loaded FITS with WCS" << endl;

		m_pFITSImage = pFITS;
		int w = m_pFITSImage->getWidth();
		int h = m_pFITSImage->getHeight();
		m_boundingRect = QRectF(0, 0, w, h);

		m_pCoordinates = m_pFITSImage->getCoordinateInterface();

		setCenterPoint((*m_pCoordinates)(IntVector2D(w/2, h/2)));

		m_imageCenterPoint = m_centerPoint;
		m_fitsFileName = fname;
	}
	else if (pFITS->loadRectCoordinates(fname.toStdString()))
	{
		clear();

		cout << "Loaded FITS without WCS" << endl;

		m_pFITSImage = pFITS;
		int w = m_pFITSImage->getWidth();
		int h = m_pFITSImage->getHeight();
		m_boundingRect = QRectF(0, 0, w, h);

		m_pCoordinates = m_pFITSImage->getCoordinateInterface();

		setCenterPoint(Vector2D<double>(0, 0));

		m_imageCenterPoint = m_centerPoint;
		m_fitsFileName = fname;
	}
	else
	{
		cout << "Unable to load FITS: " << pFITS->getErrorString() << endl;
		delete pFITS;
		return false;
	}

	return true;
}

void FITSItem::clear()
{
	if (m_pFITSImage)
		delete m_pFITSImage;
	m_pFITSImage = 0;

	if (m_pPixmap)
		delete m_pPixmap;
	m_pPixmap = 0;

	m_min = 0;
	m_max = 1;
	m_pCoordinates = 0;
	m_fitsFileName = QString();

	prepareGeometryChange();
}

void FITSItem::setCenterPoint(Vector2D<double> centerPoint)  
{
	if (m_pFITSImage == 0 || m_pCoordinates == 0)
		return;

	int w = m_pFITSImage->getWidth();
	int h = m_pFITSImage->getHeight();

	Vector2D<double> P1(0.5, 0.5);
	Vector2D<double> P2((double)w-1.0+0.5, 0.5);
	Vector2D<double> P3(0.5, (double)h-1.0+0.5);

	Vector2D<double> p1 = (*m_pCoordinates)(IntVector2D(0,0));
	Vector2D<double> p2 = (*m_pCoordinates)(IntVector2D(w-1,0));
	Vector2D<double> p3 = (*m_pCoordinates)(IntVector2D(0,h-1));

	p1 = subtractPosition(p1, centerPoint);
	p2 = subtractPosition(p2, centerPoint);
	p3 = subtractPosition(p3, centerPoint);

	p1 /= ANGLE_ARCSEC;
	p2 /= ANGLE_ARCSEC;
	p3 /= ANGLE_ARCSEC;

	Vector2D<double> d1 = p1-p3;
	Vector2D<double> d2 = p2-p3;

	Vector2D<double> D1 = P1-P3;
	Vector2D<double> D2 = P2-P3;

	double denom = D1.getX()*D2.getY()-D2.getX()*D1.getY();
	double a = 1.0/denom*(d1.getX()*D2.getY()-d2.getX()*D1.getY());
	double b = 1.0/denom*(-d1.getX()*D2.getX()+d2.getX()*D1.getX());
	double c = 1.0/denom*(d1.getY()*D2.getY()-d2.getY()*D1.getY());
	double d = 1.0/denom*(-d1.getY()*D2.getX()+d2.getY()*D1.getX());

	double e = p1.getX()-(a*P1.getX()+b*P1.getY());
	double f = p1.getY()-(c*P1.getX()+d*P1.getY());

	setTransform(QTransform(a,c,b,d,e,f), false); // Warning: b and c reversed!
	updatePointTransforms();

	m_centerPoint = centerPoint;
}

QRectF FITSItem::boundingRect() const
{
	return m_boundingRect;
}

void FITSItem::setMinMax(double min, double max)
{
	m_min = min;
	m_max = max;

	if (m_pPixmap)
	{
		delete m_pPixmap;
		m_pPixmap = 0;
	}

	update(boundingRect());
}

void FITSItem::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget)
{
	if (m_pFITSImage == 0)
		return;

	int w = m_pFITSImage->getWidth();
	int h = m_pFITSImage->getHeight();

	if (m_pPixmap == 0)
	{
		// create pixmap

		QImage image(w, h, QImage::Format_Indexed8);

		for (int i = 0 ; i < 256 ; i++)
			image.setColor(i, qRgb(i,i,i));

		double minValue = m_min;
		double maxValue = m_max;
		double denom = maxValue-minValue;
		if (denom == 0)
			denom = 1.0;

		for (int y = 0 ; y < h ; y++)
		{
			for (int x = 0 ; x < w ; x++)
			{
				double value = m_pFITSImage->getPixel(x, y);
				double pixelValue = ((value-minValue)/denom)*255.0;

				if (pixelValue > 255.0)
					pixelValue = 255.0;
				if (pixelValue < 0)
					pixelValue = 0;
				
				int rgbValue = (int)(pixelValue + 0.5);
				if (rgbValue > 255)
					rgbValue = 255;
				else if (rgbValue < 0)
					rgbValue = 0;

				image.setPixel(x, y, rgbValue);
			}
		}

		m_pPixmap = new QPixmap();
		*m_pPixmap = QPixmap::fromImage(image);
	}

	painter->drawPixmap(QPointF(0, 0), *m_pPixmap);

	//painter->setPen(Qt::blue);
	//painter->drawRect(QRectF(QPointF(0, 0), QPointF(w,h)));
}

Vector2D<double> FITSItem::subtractPosition(Vector2D<double> point, Vector2D<double> center)
{
	double a = center.getX();
	double d = center.getY();

	Vector2D<double> pos = point;

	double A = pos.getX();
	double D = pos.getY();

	double x2 = COS(d)*COS(D)*COS(A-a)+SIN(D)*SIN(d);
	double y2 = COS(D)*SIN(A-a);
	double z2 = -COS(D)*SIN(d)*COS(A-a)+SIN(D)*COS(d);

	double D2 = ASIN(z2);

	double xproj = x2/COS(D2);
	double yproj = y2/COS(D2);

	double A2 = ACOS(xproj);
	if (yproj < 0)
		A2 = -A2;

	return  Vector2D<double>(A2,D2);
}

void FITSItem::mouseDoubleClickEvent(QGraphicsSceneMouseEvent *event)
{
	if (GraleEditor::getInstance()->getCurrentTool() != GraleEditor::CenterTool)
		return;

	QPointF position = event->pos();

	Vector2D<double> coords = (*m_pCoordinates)(IntVector2D((int)(position.x()), (int)(position.y())));
	
	setCenterPoint(coords);

	emit centerChanged();
}

void FITSItem::mousePressEvent(QGraphicsSceneMouseEvent *event)
{
	if (GraleEditor::getInstance()->getCurrentTool() == GraleEditor::PointMatchTool)
	{
		double x = event->pos().x();
		double y = event->pos().y();
		bool ok = true;

		if (event->button() == Qt::RightButton)
		{
			QPointF p = mapToScene(QPointF(x, y));
			MatchPointDialog dlg(p.x(), p.y(), GraleEditor::getInstance());

			if (dlg.exec() != QDialog::Accepted)
				ok = false;
			else
			{
				x = dlg.getX();
				y = dlg.getY();

				p = mapFromScene(QPointF(x, y));

				x = p.x();
				y = p.y();
			}
		}

		if (ok)
		{
			PointMatchItem *pPoint = new PointMatchItem(this);
			LayerItem *pLayer = (LayerItem *)parentItem();

			pPoint->setPos(x, y);
			pPoint->setScaleFactor(pLayer->getPointScaleFactor());
		}
	}
	else
		QGraphicsItem::mousePressEvent(event);
}

void FITSItem::mouseReleaseEvent(QGraphicsSceneMouseEvent *event)
{
	if (GraleEditor::getInstance()->getCurrentTool() == GraleEditor::PointMatchTool)
	{
		// nothing to do
	}	
	else
		QGraphicsItem::mouseReleaseEvent(event);
}

void FITSItem::updatePointTransforms()
{
}
