/*

  This file is a part of GRALEEDITOR, a program to help create input data
  for gravitational lens inversion using GRALE.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "graleeditor.h"
#include "editorwidget.h"
#include "editorscene.h"
#include "treewidget.h"
#include <QMenuBar>
#include <QMenu>
#include <QAction>
#include <QSplitter>
#include <QActionGroup>
#include <QMessageBox>
#include <QFrame>
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QGridLayout>
#include <QPushButton>
#include <QKeyEvent>
#include <QFileDialog>
#include <QFileInfo>
#include <QToolBar>
#include <QColor>
#include <QBrush>
#include <QPixmap>
#include <QInputDialog>
#include <QLabel>
#include <QLineEdit>
#include <QDoubleValidator>
#include <QSettings>
#include <QTimer>
#include <grale/constants.h>
#include <grale/triangulation.h>
#include <grale/polygon2d.h>
#include <grale/imagesdata.h>
#include "fitsitem.h"
#include "pointitem.h"
#include "imageitem.h"
#include "triangleitem.h"
#include "imagelayertreeitem.h"
#include "pointslayertreeitem.h"
#include "gridinputdialog.h"

#include <limits>
#include <iostream>
#include <vector>
#include <assert.h>

using namespace grale;
using namespace std;

const char kWinGeom[] = "mainWindowGeometry";
const char kWinState[] = "mainWindowState";
const char kExportBitmapWidth[] = "mainExportBitmapWidth";
const char kDefaultPointSize[] = "mainDefaultPointSize";
const char kDefaultMatchPointSize[] = "mainDefaultMatchPointSize";
const char kAxesShown[] = "mainAxesShown";

GraleEditor *GraleEditor::m_pInstance = 0;

GraleEditor::GraleEditor(const QVector<QString> &args)
{
	m_pInstance = this;

	QSplitter *pSplitter = new QSplitter(Qt::Horizontal, this);
	setCentralWidget(pSplitter);

	m_pViewSplitter = new QSplitter(Qt::Vertical, pSplitter);

	EditorWidget *pEditWidget = new EditorWidget(m_pViewSplitter);	
	m_editorWidgets.append(pEditWidget);

	m_pScene = new EditorScene();
	pEditWidget->setScene(m_pScene);

	QFrame *pRightFrame = new QFrame(pSplitter);

	QVBoxLayout *pVLayout = new QVBoxLayout(pRightFrame);

	m_pLayersWidget = new TreeWidget(pRightFrame);
	QStringList labels;

	labels.append("Layers");
	labels.append("Value");
	m_pLayersWidget->setHeaderLabels(labels);
	pVLayout->addWidget(m_pLayersWidget);

	QFrame *pFrame = new QFrame(pRightFrame);
	pVLayout->addWidget(pFrame);

	QGridLayout *pGridLayout = new QGridLayout(pFrame);
	
	QLabel *pLabel1 = new QLabel("Default point size", pFrame);
	QLabel *pLabel2 = new QLabel("Default match point size", pFrame);
	m_pDefaultPointSizeEdit = new QLineEdit(pFrame);
	m_pDefaultMatchPointSizeEdit = new QLineEdit(pFrame);
	m_pDefaultPointSizeEdit->setValidator(new QDoubleValidator(0.00001, 100.0, 10, m_pDefaultPointSizeEdit));
	m_pDefaultMatchPointSizeEdit->setValidator(new QDoubleValidator(0.00001,100.0,10, m_pDefaultMatchPointSizeEdit));

	connect(m_pDefaultPointSizeEdit, SIGNAL(editingFinished()), this, SLOT(onDefaultPointSizeChanged()));
	connect(m_pDefaultMatchPointSizeEdit, SIGNAL(editingFinished()), this, SLOT(onDefaultMatchPointSizeChanged()));

	pGridLayout->addWidget(pLabel1, 0, 0);
	pGridLayout->addWidget(pLabel2, 1, 0);
	pGridLayout->addWidget(m_pDefaultPointSizeEdit, 0, 1);
	pGridLayout->addWidget(m_pDefaultMatchPointSizeEdit, 1, 1);
	pGridLayout->setColumnMinimumWidth(1, 100);

	pFrame = new QFrame(pRightFrame);
	pVLayout->addWidget(pFrame);

	QGridLayout *pHBoxLayout = new QGridLayout(pFrame);

	QPushButton *pAddImgLayer = new QPushButton("Add image layer");
	QPushButton *pAddPointsLayer = new QPushButton("Add points layer");
	QPushButton *pUpButton = new QPushButton("Move up", pFrame);
	QPushButton *pDownButton = new QPushButton("Move down", pFrame);
	pHBoxLayout->addWidget(pAddImgLayer, 0, 0);
	pHBoxLayout->addWidget(pAddPointsLayer, 0, 1);
	pHBoxLayout->addWidget(pUpButton, 1, 0);
	pHBoxLayout->addWidget(pDownButton, 1, 1);

	connect(m_pLayersWidget, SIGNAL(itemDoubleClicked(QTreeWidgetItem *, int)),
		this, SLOT(onLayerItemDoubleClicked(QTreeWidgetItem *, int)));
	connect(m_pLayersWidget, SIGNAL(itemChanged(QTreeWidgetItem *, int)),
		this, SLOT(onLayerItemChanged(QTreeWidgetItem *, int)));
	connect(pUpButton, SIGNAL(clicked()), this, SLOT(onLayerUp()));
	connect(pDownButton, SIGNAL(clicked()), this, SLOT(onLayerDown()));
	connect(pAddPointsLayer, &QPushButton::clicked, this, &GraleEditor::onAddPointsLayer);
	connect(pAddImgLayer, &QPushButton::clicked, this, &GraleEditor::onAddImageLayer);

	createMenus();
	statusBar();

	show();

	QSettings settings;
	auto winGeom = settings.value(kWinGeom);
	auto winState = settings.value(kWinState);
	restoreGeometry(winGeom.toByteArray());
	restoreState(winState.toByteArray());

	m_lastExportBitmapWidth = settings.value(kExportBitmapWidth, 800).toInt();

	m_pScene->setDefaultPointScaleFactor(settings.value(kDefaultPointSize, 0.05).toDouble());
	m_pScene->setDefaultMatchPointScaleFactor(settings.value(kDefaultMatchPointSize, 0.1).toDouble());
	m_pDefaultPointSizeEdit->setText(QString::asprintf("%.10f", m_pScene->getDefaultPointScaleFactor()));
	m_pDefaultMatchPointSizeEdit->setText(QString::asprintf("%.10f", m_pScene->getDefaultMatchPointScaleFactor()));

	m_lastSetPointSize = m_pScene->getDefaultPointScaleFactor();
	m_lastSetMatchPointSize = m_pScene->getDefaultMatchPointScaleFactor();

	m_pScene->setCoordinateItem(settings.value(kAxesShown, true).toBool());

	if (args.size() > 0)
	{
		m_args = args;

		QTimer *pLoadArgsTimer = new QTimer(this);
		pLoadArgsTimer->setSingleShot(true);
		pLoadArgsTimer->setInterval(0);
		QObject::connect(pLoadArgsTimer, &QTimer::timeout, this, &GraleEditor::onProcessArgs);
		pLoadArgsTimer->start();
	}
}

GraleEditor::~GraleEditor()
{
}

void GraleEditor::onProcessArgs()
{
	for (auto a : m_args)
	{
		QString lowerCase = a.toLower();
		if (lowerCase.endsWith(".ge"))
		{
			onNewInternal();
			onLoadInternal(a);
		}
		else if (lowerCase.endsWith(".jpg") || lowerCase.endsWith(".png") || lowerCase.endsWith(".jpeg") ||
		    lowerCase.endsWith(".fits"))
		{
			onAddImageLayer();
			int layerIndex = m_pScene->getNumberOfLayers()-1; // The last one was added
			LayerItem *pNewLayer = m_pScene->getLayerItem(layerIndex);

			const int treeCount = m_pLayersWidget->topLevelItemCount();
			for (int i = 0 ; i < treeCount ; i++)
			{
				TreeItemBase *pItem = static_cast<TreeItemBase *>(m_pLayersWidget->topLevelItem(i));
				if (pItem->getType() == TreeItemBase::Layer)
				{
					LayerTreeItem *pLayerItem = static_cast<LayerTreeItem *>(pItem);
					if (pLayerItem->getLayer() == pNewLayer)
					{
						ImageLayerTreeItem *pImageTreeItem = static_cast<ImageLayerTreeItem *>(pLayerItem);
						pImageTreeItem->loadImage(a);
						break;
					}
				}
			}
		}
		else if (lowerCase.endsWith(".imgdata"))
		{
			onImportImagesDataInternal(a);
		}
		else if (a == "--noaxes" || a == "--noaxis")
		{
			m_pScene->setCoordinateItem(false);
		}
		else if (a == "--axes" || a == "--axis")
		{
			m_pScene->setCoordinateItem(true);
		}
		else
			cerr << "Ignoring command line argument '" << a.toStdString() << "'" << endl;
	}
}

void GraleEditor::createMenus()
{
	QAction *pAction;

	QMenu *pFileMenu = menuBar()->addMenu("File");

	pAction = new QAction("New", this);
	connect(pAction, SIGNAL(triggered()), this, SLOT(onNew()));
	pFileMenu->addAction(pAction);

	pAction = new QAction("Load", this);
	connect(pAction, SIGNAL(triggered()), this, SLOT(onLoad()));
	pFileMenu->addAction(pAction);

	pAction = new QAction("Save", this);
	connect(pAction, SIGNAL(triggered()), this, SLOT(onSave()));
	pFileMenu->addAction(pAction);

	pFileMenu->addSeparator();

	pAction = new QAction("Import images data", this);
	connect(pAction, SIGNAL(triggered()), this, SLOT(onImportImagesData()));
	pFileMenu->addAction(pAction);

	pAction = new QAction("Export images data", this);
	connect(pAction, SIGNAL(triggered()), this, SLOT(onExportImagesData()));
	pFileMenu->addAction(pAction);

	QMenu *pExportOptions = pFileMenu->addMenu("Export options");
	pAction = new QAction("Export triangles", this);
	pAction->setCheckable(true);
	pAction->setChecked(true);
	pExportOptions->addAction(pAction);
	m_pActionTriangles = pAction; 

	pAction = new QAction("Export intensities", this);
	pAction->setCheckable(true);
	pAction->setChecked(false);
	pExportOptions->addAction(pAction);
	m_pActionIntensities = pAction; 

	pAction = new QAction("Export group info", this);
	pAction->setCheckable(true);
	pAction->setChecked(true);
	pExportOptions->addAction(pAction);
	m_pActionGroups = pAction;

	pAction = new QAction("Export time delays", this);
	pAction->setCheckable(true);
	pAction->setChecked(true);
	pExportOptions->addAction(pAction);
	m_pActionTimeDelays = pAction;

	pFileMenu->addSeparator();

	pAction = new QAction("Save view to bitmap", this);
	connect(pAction, SIGNAL(triggered()), this, SLOT(onExportBitmap()));
	pFileMenu->addAction(pAction);

	pAction = new QAction("Save area to bitmap", this);
	connect(pAction, SIGNAL(triggered()), this, SLOT(onAreaExportBitmap()));
	pFileMenu->addAction(pAction);

	QMenu *pEditMenu = menuBar()->addMenu("Edit");

	pAction = new QAction("Copy", this);
	pAction->setShortcuts(QKeySequence::Copy);
	connect(pAction, SIGNAL(triggered()), this, SLOT(onCopy()));
	pEditMenu->addAction(pAction);

	pAction = new QAction("Paste", this);
	pAction->setShortcuts(QKeySequence::Paste);
	connect(pAction, SIGNAL(triggered()), this, SLOT(onPaste()));
	pEditMenu->addAction(pAction);

	pEditMenu->addSeparator();

	pAction = new QAction("Set point size", this);
	connect(pAction, SIGNAL(triggered()), this, SLOT(onSetPointSize()));
	pEditMenu->addAction(pAction);

	pAction = new QAction("Set match point size", this);
	connect(pAction, SIGNAL(triggered()), this, SLOT(onSetMatchPointSize()));
	pEditMenu->addAction(pAction);

	QMenu *pViewMenu = menuBar()->addMenu("View");

	pAction = new QAction("Add view", this);
	connect(pAction, SIGNAL(triggered()), this, SLOT(onAddView()));
	pViewMenu->addAction(pAction);

	pAction = new QAction("Delete last view", this);
	connect(pAction, SIGNAL(triggered()), this, SLOT(onDeleteView()));
	pViewMenu->addAction(pAction);

	pAction = new QAction("Toggle coordinate axes", this);
	connect(pAction, SIGNAL(triggered()), m_pScene, SLOT(toggleCoordinateItem()));
	pViewMenu->addAction(pAction);

	pViewMenu->addSeparator();

	QActionGroup *pOrientationGroup = new QActionGroup(this);
	m_pXLeftAction = new QAction("X Axis points left", pOrientationGroup);
	m_pXLeftAction->setCheckable(true);
	m_pXLeftAction->setChecked(true);
	pViewMenu->addAction(m_pXLeftAction);
	m_pXRightAction = new QAction("X Axis points right", pOrientationGroup);
	m_pXRightAction->setCheckable(true);
	pViewMenu->addAction(m_pXRightAction);

	connect(pOrientationGroup, SIGNAL(triggered(QAction *)), this, SLOT(onOrientationChanged(QAction *)));

	QMenu *pLayerMenu = menuBar()->addMenu("Layer");

	pAction = new QAction("Add image layer", this);
	connect(pAction, SIGNAL(triggered()), this, SLOT(onAddImageLayer()));
	pLayerMenu->addAction(pAction);

	pAction = new QAction("Add points layer", this);
	connect(pAction, SIGNAL(triggered()), this, SLOT(onAddPointsLayer()));
	pLayerMenu->addAction(pAction);

	QMenu *pToolMenu = menuBar()->addMenu("Tool");
	QToolBar *pToolToolBar = addToolBar("Tools");
	pToolToolBar->setObjectName("Toolbar");

	QActionGroup *pToolGroup = new QActionGroup(this);
	m_pSelectAction = new QAction("Select/Move", pToolGroup);
	m_pSelectAction->setCheckable(true);
	pToolMenu->addAction(m_pSelectAction);
	pToolToolBar->addAction(m_pSelectAction);
	m_pAddPointAction = new QAction("Add point", pToolGroup);
	m_pAddPointAction->setCheckable(true);
	pToolMenu->addAction(m_pAddPointAction);
	pToolToolBar->addAction(m_pAddPointAction);
	m_pCenterAction = new QAction("Center image", pToolGroup);
	m_pCenterAction->setCheckable(true);
	pToolMenu->addAction(m_pCenterAction);
	pToolToolBar->addAction(m_pCenterAction);
	m_pGroupNumberAction = new QAction("Group number", pToolGroup);
	m_pGroupNumberAction->setCheckable(true);
	pToolMenu->addAction(m_pGroupNumberAction);
	pToolToolBar->addAction(m_pGroupNumberAction);
	m_pPointInfoAction = new QAction("Edit point info", pToolGroup);
	m_pPointInfoAction->setCheckable(true);
	pToolMenu->addAction(m_pPointInfoAction);
	pToolToolBar->addAction(m_pPointInfoAction);
	m_pPointMatchTool = new QAction("Point match tool", pToolGroup);
	m_pPointMatchTool->setCheckable(true);
	pToolMenu->addAction(m_pPointMatchTool);
	pToolToolBar->addAction(m_pPointMatchTool);
	m_pHullToolAction = new QAction("Draw hull", pToolGroup);
	m_pHullToolAction->setCheckable(true);
	pToolMenu->addAction(m_pHullToolAction);
	pToolToolBar->addAction(m_pHullToolAction);
	m_pSelectAction->setChecked(true);

	connect(pToolGroup, SIGNAL(triggered(QAction *)), this, SLOT(onToolChanged(QAction *)));

	pToolMenu->addSeparator();
	pAction = new QAction("Add one triangle", this);
	connect(pAction, SIGNAL(triggered()), this, SLOT(onAddOneTriangle()));
	pToolMenu->addAction(pAction);

	pAction = new QAction("Triangulate", this);
	connect(pAction, SIGNAL(triggered()), this, SLOT(onTriangulate()));
	pToolMenu->addAction(pAction);

	pAction = new QAction("Add grid", this);
	connect(pAction, SIGNAL(triggered()), this, SLOT(onAddGrid()));
	pToolMenu->addAction(pAction);

	QMenu *pHelpMenu = menuBar()->addMenu("Help");
}

void GraleEditor::onAddImageLayer()
{
	int newLayer = m_pScene->addLayer(LayerItem::Image);

	m_pScene->setCurrentLayer(newLayer);

	LayerItem *pLayer = m_pScene->getLayerItem(newLayer);
	ImageLayerTreeItem *pTreeItem = new ImageLayerTreeItem(pLayer, m_pLayersWidget);

	connect(pLayer, SIGNAL(signalGroupNumberChanged()), pTreeItem, SLOT(updateImageProperties()));

	toolCheck();
	displayActiveLayer();
}

void GraleEditor::onAddPointsLayer()
{
	int newLayer = m_pScene->addLayer(LayerItem::Points);

	m_pScene->setCurrentLayer(newLayer);

	LayerItem *pLayer = m_pScene->getLayerItem(newLayer);
	PointsLayerTreeItem *pTreeItem = new PointsLayerTreeItem(pLayer, m_pLayersWidget);

	connect(pLayer, SIGNAL(signalGroupNumberChanged()), pTreeItem, SLOT(updateProperties()));

	toolCheck();
	displayActiveLayer();
}

void GraleEditor::onLayerItemDoubleClicked(QTreeWidgetItem *item, int column)
{
	TreeItemBase *pItem = (TreeItemBase *)item;
	
	if (pItem->getType() == TreeItemBase::LoadImage)
	{
		ImageLayerTreeItem *pImageItem = (ImageLayerTreeItem *)(pItem->parent());

		pImageItem->loadImage();
	}
	else if (pItem->getType() == TreeItemBase::Layer)
	{
		int idx = m_pLayersWidget->indexOfTopLevelItem(pItem);
		
		// This is something nasty to prevent the item from collapsing/expanding when
		// double clicked
		item->setExpanded(!item->isExpanded()); 

		m_pScene->setCurrentLayer(idx);

		toolCheck();
		displayActiveLayer();
	}
	else if (pItem->getType() == TreeItemBase::MatchFITS)
	{
		ImageLayerTreeItem *pImageTreeItem = (ImageLayerTreeItem *)(pItem->parent());

		ImageItem *pImageItem = pImageTreeItem->getImageItem();
		FITSItem *pFITSItem = 0;

		for (int i = 0 ; pFITSItem == 0 && i < m_pLayersWidget->topLevelItemCount() ; i++)
		{
			LayerTreeItem *pLayerTreeItem = (LayerTreeItem *)m_pLayersWidget->topLevelItem(i);
			
			if (pLayerTreeItem->getLayer()->getLayerType() == LayerItem::Image && 
			    pLayerTreeItem->getLayer()->isVisible())
			{
				ImageLayerTreeItem *pTreeItem = (ImageLayerTreeItem *)pLayerTreeItem;

				pFITSItem = pTreeItem->getFITSItem();
				if (pFITSItem)
				{
					if (!pFITSItem->isImageLoaded())
						pFITSItem = 0;
				}
			}
		}

		if (pFITSItem == 0)
		{
			QMessageBox::warning(this, "Error", "No FITS layer found to match image to");
			return;
		}

		pImageItem->matchFITSItem(pFITSItem);
	}
}

void GraleEditor::onLayerItemChanged(QTreeWidgetItem *item, int column)
{
	TreeItemBase *pItem = (TreeItemBase *)item;
	
	if (pItem->getType() == TreeItemBase::Layer && column == 0)
	{
		LayerTreeItem *pTreeItem = (LayerTreeItem *)pItem;
		LayerItem *pLayer = pTreeItem->getLayer();

		if (pItem->checkState(0) == Qt::Checked)
			pLayer->setVisible(true);
		else
			pLayer->setVisible(false);
		return;
	}

	if (pItem->getType() == TreeItemBase::PointScale && column == 1)
	{
		LayerTreeItem *pTreeItem = (LayerTreeItem *)(pItem->parent());

		bool ok;
		double oldScale = pTreeItem->getLayer()->getPointScaleFactor();
		double newScale = pItem->text(1).toDouble(&ok);

		if (ok)
			pTreeItem->getLayer()->setPointScaleFactor(newScale);

		if (pTreeItem->getLayer()->getLayerType() == LayerItem::Points)
		{
			PointsLayerTreeItem *pPointItem = (PointsLayerTreeItem *)pTreeItem;
			pPointItem->updateProperties();
		}
		else // image layer
		{
			ImageLayerTreeItem *pImageItem = (ImageLayerTreeItem *)pTreeItem;
			pImageItem->updateImageProperties();
		}

		return;
	}

	if (pItem->getType() == TreeItemBase::GroupNumber && column == 1)
	{
		LayerTreeItem *pTreeItem = (LayerTreeItem *)(pItem->parent());

		bool ok;
		int newGroup = pItem->text(1).toInt(&ok, 10);

		pTreeItem->getLayer()->setGroupNumber(newGroup);
		// update is not needed due to signal/slot connection

		return;
	}

	if (!((pItem->getType() == TreeItemBase::XCenter) || (pItem->getType() == TreeItemBase::YCenter)
	   ||(pItem->getType() == TreeItemBase::MinValue) || (pItem->getType() == TreeItemBase::MaxValue)) )
		return;

	if (column != 1)
		return;

	ImageLayerTreeItem *pImageItem = (ImageLayerTreeItem *)(pItem->parent());

	Vector2D<double> curCenter = pImageItem->getFITSItem()->getCenterPoint();
	Vector2D<double> imageCenter = pImageItem->getFITSItem()->getImageCenterPoint();
	double newX, newY;
	double min, max;
	double newMin, newMax;
	bool ok;

	pImageItem->getFITSItem()->getMinMax(min, max);

	newX = curCenter.getX();
	newY = curCenter.getY();

	newMin = min;
	newMax = max;

	if (pItem->getType() == TreeItemBase::XCenter)
	{
		newX = pItem->text(1).toDouble(&ok);
		newX *= CONST_PI/180.0;
	}
	else if (pItem->getType() == TreeItemBase::YCenter)
	{
		newY = pItem->text(1).toDouble(&ok);
		newY *= CONST_PI/180.0;
	}
	else if (pItem->getType() == TreeItemBase::MinValue)
	{
		newMin = pItem->text(1).toDouble(&ok);
	}
	else if (pItem->getType() == TreeItemBase::MaxValue)
	{
		newMax = pItem->text(1).toDouble(&ok);
	}

	if (!ok)
	{
		pImageItem->updateImageProperties();
		return;
	}

	Vector2D<double> newCenter(newX, newY);
	Vector2D<double> diff = newCenter-imageCenter;

	diff /= ANGLE_ARCMIN;

	if (diff.getLength() > 20.0)
	{
		QMessageBox::warning(this, "Error", "Specified center too far away from image center");
	}
	else
	{
		pImageItem->getFITSItem()->setCenterPoint(newCenter);
		pImageItem->getFITSItem()->setMinMax(newMin, newMax);
	}

	pImageItem->updateImageProperties();
}

GraleEditor::Tool GraleEditor::getCurrentTool() const
{
	if (m_pSelectAction->isChecked())
		return SelectTool;
	if (m_pAddPointAction->isChecked())
		return AddPointTool;
	if (m_pCenterAction->isChecked())
		return CenterTool;
	if (m_pGroupNumberAction->isChecked())
		return GroupNumberTool;
	if (m_pPointInfoAction->isChecked())
		return PointInfoTool;
	if (m_pPointMatchTool->isChecked())
		return PointMatchTool;
	if (m_pHullToolAction->isChecked())
		return HullTool;
	return InvalidTool;
}

void GraleEditor::onLayerUp()
{
	QList<QTreeWidgetItem *> items = m_pLayersWidget->selectedItems();

	if (items.size() != 1)
		return;

	QTreeWidgetItem *pItem = items.at(0);

	int idx = m_pLayersWidget->indexOfTopLevelItem(pItem);

	if (idx < 1)
		return;

	pItem = m_pLayersWidget->takeTopLevelItem(idx);
	m_pLayersWidget->insertTopLevelItem(idx-1, pItem);
	pItem->setExpanded(true);

	for (int i = 0 ; i < m_pLayersWidget->topLevelItemCount() ; i++)
		m_pLayersWidget->topLevelItem(i)->setSelected(false);
	pItem->setSelected(true);

	m_pScene->decreaseLayerIndex(idx);

	toolCheck();
	displayActiveLayer();
}

void GraleEditor::onLayerDown()
{
	QList<QTreeWidgetItem *> items = m_pLayersWidget->selectedItems();

	if (items.size() != 1)
		return;

	QTreeWidgetItem *pItem = items.at(0);

	int idx = m_pLayersWidget->indexOfTopLevelItem(pItem);

	if (idx >= m_pLayersWidget->topLevelItemCount()-1)
		return;

	pItem = m_pLayersWidget->takeTopLevelItem(idx);
	m_pLayersWidget->insertTopLevelItem(idx+1, pItem);
	pItem->setExpanded(true);

	for (int i = 0 ; i < m_pLayersWidget->topLevelItemCount() ; i++)
		m_pLayersWidget->topLevelItem(i)->setSelected(false);
	pItem->setSelected(true);

	m_pScene->increaseLayerIndex(idx);

	toolCheck();
	displayActiveLayer();
}

void GraleEditor::displayActiveLayer()
{
	int num = m_pLayersWidget->topLevelItemCount();
	QBrush inactiveBrush(Qt::transparent);
	QBrush activeBrush(Qt::green);

	for (int i = 0 ; i < num ; i++)
	{
		LayerTreeItem *pLayerItem = (LayerTreeItem *)m_pLayersWidget->topLevelItem(i);

		if (pLayerItem->getLayer()->isLayerEnabled())
		{
			pLayerItem->setBackground(0, activeBrush);
			pLayerItem->setBackground(1, activeBrush);
			pLayerItem->setText(1,"ACTIVE");
		}
		else
		{
			pLayerItem->setBackground(0, inactiveBrush);
			pLayerItem->setBackground(1, inactiveBrush);
			pLayerItem->setText(1,"");
		}
	}
}

void GraleEditor::onAddOneTriangle()
{
	int idx = m_pScene->getCurrentLayer();
	LayerItem *pLayer = m_pScene->getLayerItem(idx);

	if (pLayer == 0)
		return;

	if (pLayer->getLayerType() != LayerItem::Points)
	{
		QMessageBox::warning(this, "Error", "Active layer is not a points layer");
		return;
	}

	QList<PointItem *> pointList = pLayer->getSelectedPoints();

	if (pointList.size() != 3)
	{
		QMessageBox::warning(this, "Error", "Exactly three points must be selected for this operation");
		return;
	}

	TriangleItem *pTriangle = new TriangleItem(pointList.at(0), pointList.at(1), pointList.at(2), pLayer);

	toolCheck();
}

void GraleEditor::onTriangulate()
{
	int idx = m_pScene->getCurrentLayer();
	LayerItem *pLayer = m_pScene->getLayerItem(idx);

	if (pLayer == 0)
		return;

	if (pLayer->getLayerType() != LayerItem::Points)
	{
		QMessageBox::warning(this, "Error", "Active layer is not a points layer");
		return;
	}


	QList<PointItem *> pointList = pLayer->getSelectedPoints();

	if (pointList.size() < 3)
	{
		QMessageBox::warning(this, "Error", "At least three points must be selected for this operation");
		return;
	}

	std::vector<Vector2D<double> > points(pointList.size());	

	for (int i = 0 ; i < pointList.size() ; i++)
	{
		PointItem *pItem = pointList.at(i);
		QPointF pos = pItem->pos();

		points[i] = Vector2D<double>(pos.x(), pos.y());
	}

	Triangulation triangulation;

	if (!triangulation.create(points))
	{
		QMessageBox::warning(this, "Error", triangulation.getErrorString().c_str());
		return;
	}

	std::list<TriangleIndices> triangles;
	std::list<TriangleIndices>::const_iterator it;

	triangulation.getTriangles(triangles);

	for (it = triangles.begin() ; it != triangles.end() ; it++)
	{
		TriangleIndices indices = *it;

		PointItem *pPoint1 = pointList.at(indices.getIndex(0));
		PointItem *pPoint2 = pointList.at(indices.getIndex(1));
		PointItem *pPoint3 = pointList.at(indices.getIndex(2));

		TriangleItem *pTriangleItem = new TriangleItem(pPoint1, pPoint2, pPoint3, pLayer);
	}

	toolCheck();
}

void GraleEditor::onAddGrid()
{
	int idx = m_pScene->getCurrentLayer();
	LayerItem *pLayer = m_pScene->getLayerItem(idx);

	if (pLayer == 0)
		return;

	if (pLayer->getLayerType() != LayerItem::Points)
	{
		QMessageBox::warning(this, "Error", "Active layer is not a points layer");
		return;
	}

	GridInputDialog gridDialog(this);

	if (gridDialog.exec() != QDialog::Accepted)
		return;

	vector<ImagesData *> excludeImages; // we'll store a copy since we might be modifying a layer itself
	if (gridDialog.getExcludeAreas())
	{
		//cerr << "Excluding areas" << endl;
		const int numLayers = m_pScene->getNumberOfLayers();
		QList<LayerItem *> pointLayers;

		for (int i = 0 ; i < numLayers ; i++)
		{
			LayerItem *pLayer = m_pScene->getLayerItem(i);

			if (pLayer->isVisible() && pLayer->getLayerType() == LayerItem::Points)
				pointLayers.append(pLayer);
		}

		//cout << "Visible points layers: " << pointLayers.size() << endl;

		for (int i = 0 ; i < pointLayers.size() ; i++)
		{
			LayerItem *pLayer = pointLayers.at(i);
			ImagesData *pImgDat = new ImagesData();
			excludeImages.push_back(pImgDat);

			// TODO: error check
			if (!pImgDat->create(1, false, false))
				cerr << "WARNING: couldn't create image!" << endl;
			else
				cerr << "Created image" << endl;

			map<PointItem *, int> pointMap;
			QList<PointItem *> pointList = pLayer->getAllPoints();

			for (int p = 0 ; p < pointList.size() ; p++)
			{
				PointItem *pPoint = pointList.at(p);
				QPointF pos = pPoint->pos();

				// Note that for correctness we should convert the unit here, but since we're
				// only using the imgdata internally, it doesn't matter
				Vector2D<double> angularPos(pos.x(), pos.y());

				// TODO: should really check for errors here
				pointMap[pPoint] = pImgDat->addPoint(0, angularPos);
				//cout << "Added pos at " << angularPos.getX() << " " << angularPos.getY() << ": " << pointMap[pPoint] << endl;
			}

			QList<TriangleItem *> triangleList = pLayer->getAllTriangles();
			for (TriangleItem *pTriangle : triangleList)
			{
				int indices[3];

				for (int p = 0 ; p < 3 ; p++)
				{
					PointItem *pPoint = pTriangle->getPoint(p);
					indices[p] = pointMap[pPoint];
				}

				// TODO: check for errors
				//cerr << "Indices: " << indices[0] << " " << indices[1] << " " << indices[2] << endl;
				if (!pImgDat->addTriangle(0, indices[0], indices[1], indices[2]))
					cerr << "Error adding triangle: " << pImgDat->getErrorString() << endl;
			}
		}
	}

	int numX = gridDialog.getXPoints();
	int numY = gridDialog.getYPoints();
	double xCenter = gridDialog.getXCenter();
	double yCenter = gridDialog.getYCenter();
	double width = gridDialog.getWidth();
	double height = gridDialog.getHeight();

	if (!gridDialog.getExcludeAreas())
	{
		std::vector<std::vector<PointItem *> > points;

		points.resize(numY);
		for (int y = 0 ; y < numY ; y++)
		{
			double yCoord = yCenter - height/2.0 + ((double)y)*(height/((double)(numY-1)));

			points[y].resize(numX);

			for (int x = 0 ; x < numX ; x++)
			{
				double xCoord = xCenter - width/2.0 + ((double)x)*(width/((double)(numX-1)));

				PointItem *pPoint = new PointItem(pLayer);
				pPoint->setPos(xCoord, yCoord);
				pPoint->setScaleFactor(pLayer->getPointScaleFactor());

				points[y][x] = pPoint;
			}
		}

		if (gridDialog.getAddTriangles())
		{
			for (int y = 0 ; y < numY-1 ; y++)
			{
				for (int x = 0 ; x < numX-1 ; x++)
				{
					TriangleItem *pTriangle1 = new TriangleItem(points[y][x], points[y+1][x], points[y][x+1], pLayer);
					TriangleItem *pTriangle2 = new TriangleItem(points[y+1][x+1], points[y+1][x], points[y][x+1], pLayer);
				}
			}
		}
	}
	else
	{
		vector<PointItem *> points;

		// Add the borders of the images
		vector<Polygon2D<double>> polygons;
		for (ImagesData *pImgDat : excludeImages)
		{
			Polygon2D<double> poly;
			if (!pImgDat->getImageBorder(0, poly))
				cerr << "WARNING: couldn't get image border: " << pImgDat->getErrorString() << endl;
			polygons.push_back(poly);

			const Vector2D<double> *pPolyPoints = poly.getPoints();
			const int numPoints = poly.getNumberOfPoints();

			//cerr << "Got polygon with " << numPoints << " points" << endl;

			for (int p = 0 ; p < numPoints ; p++)
			{
				Vector2D<double> polyPoint = pPolyPoints[p];
				//cout << " " << polyPoint.getX() << " " << polyPoint.getY() << endl;

				PointItem *pPoint = new PointItem(pLayer);
				pPoint->setPos(polyPoint.getX(), polyPoint.getY());
				pPoint->setScaleFactor(pLayer->getPointScaleFactor());

				points.push_back(pPoint);
			}
		}

		// Add the grid points that are not inside these polygons
		for (int y = 0 ; y < numY ; y++)
		{
			double yCoord = yCenter - height/2.0 + ((double)y)*(height/((double)(numY-1)));

			for (int x = 0 ; x < numX ; x++)
			{
				double xCoord = xCenter - width/2.0 + ((double)x)*(width/((double)(numX-1)));
				bool isInside = false;

				for (Polygon2D<double> &poly : polygons)
				{
					if (poly.isInside(Vector2D<double>(xCoord, yCoord)))
					{
						isInside = true;
						break;
					}
				}

				if (!isInside)
				{
					PointItem *pPoint = new PointItem(pLayer);
					pPoint->setPos(xCoord, yCoord);
					pPoint->setScaleFactor(pLayer->getPointScaleFactor());

					points.push_back(pPoint);
				}
			}
		}

		if (gridDialog.getAddTriangles())
		{
			// Create a triangulation
			vector<Vector2D<double>> gralePoints;
			for (PointItem *pPointItem : points)
			{
				QPointF pos = pPointItem->pos();
				gralePoints.push_back(Vector2D<double>(pos.x(), pos.y()));
			}
			
			Triangulation triangulation;
			if (!triangulation.create(gralePoints))
			{
				QMessageBox::warning(this, "Error creating triangulation", triangulation.getErrorString().c_str());
			}
			else
			{
				list<TriangleIndices> triangles;
				triangulation.getTriangles(triangles);

				for (auto it = triangles.begin() ; it != triangles.end() ; it++)
				{
					TriangleIndices indices = *it;

					PointItem *p0 = points[indices.getIndex(0)];
					PointItem *p1 = points[indices.getIndex(1)];
					PointItem *p2 = points[indices.getIndex(2)];
					PointItem *ppPoints[] = { p0, p1, p2 };

					double xc = 0, yc = 0;
					for (int i = 0 ; i < 3 ; i++)
					{
						QPointF pos = ppPoints[i]->pos();
						xc += pos.x();
						yc += pos.y();
					}
					xc /= 3.0;
					yc /= 3.0;

					bool isInside = false;

					for (Polygon2D<double> &poly : polygons)
					{
						if (poly.isInside(Vector2D<double>(xc, yc)))
						{
							isInside = true;
							break;
						}
					}

					if (!isInside)
					{
						TriangleItem *pTriangleItem = new TriangleItem(ppPoints[0], ppPoints[1], ppPoints[2], pLayer);
					}
				}
			}
		}
	}

	toolCheck();

	for (ImagesData *pImg : excludeImages)
		delete pImg;
}

void GraleEditor::keyPressEvent(QKeyEvent *event)
{
	if (event->key() == Qt::Key_Delete)
	{
		event->accept();
	}
}

void GraleEditor::keyReleaseEvent(QKeyEvent *event)
{
	if (event->key() == Qt::Key_Delete)
	{
		event->accept();

		QList<QTreeWidgetItem *> items = m_pLayersWidget->selectedItems();

		if (items.size() != 1)
			return;

		QTreeWidgetItem *pItem = items.at(0);
		int idx = m_pLayersWidget->indexOfTopLevelItem(pItem);

		QMessageBox::StandardButton result;

		result = QMessageBox::question(this, "Confirm", "Delete selected layer?", QMessageBox::Yes|QMessageBox::No, QMessageBox::No);

		if (result == QMessageBox::Yes)
		{
			pItem = m_pLayersWidget->takeTopLevelItem(idx);
			delete pItem;

			m_pScene->deleteLayer(idx);

			toolCheck();
			displayActiveLayer();
		}
	}
}

void GraleEditor::onImportImagesData()
{
	QString fileName = QFileDialog::getOpenFileName(this, "Open images data", "", "Images data files (*.imgdata)");
		
	if (fileName.isNull())
		return;

	onImportImagesDataInternal(fileName);
}

void GraleEditor::onImportImagesDataInternal(const QString &fileName)
{
	QList<LayerItem *> newLayers;

	if (!m_pScene->importImagesData(fileName, newLayers))
	{
		QString msg = "Couldn't import images data file";

		msg += m_pScene->getErrorString().c_str();
		QMessageBox::warning(this, "Error", msg);
		return;
	}

	for (int i = 0 ; i < newLayers.size() ; i++)
	{
		LayerItem *pLayer = newLayers.at(i);
		PointsLayerTreeItem *pTreeItem = new PointsLayerTreeItem(pLayer, m_pLayersWidget);

		connect(pLayer, SIGNAL(signalGroupNumberChanged()), pTreeItem, SLOT(updateProperties()));
	}

	toolCheck();
	displayActiveLayer();
}

void GraleEditor::onExportImagesData()
{
	int numLayers = m_pScene->getNumberOfLayers();
	QList<LayerItem *> pointLayers;

	for (int i = 0 ; i < numLayers ; i++)
	{
		LayerItem *pLayer = m_pScene->getLayerItem(i);

		if (pLayer->isVisible() && pLayer->getLayerType() == LayerItem::Points)
			pointLayers.append(pLayer);
	}

	if (pointLayers.size() == 0)
	{
		QMessageBox::warning(this, "Error", "No visible point layers to export");
		return;
	}

	char str[1024];
	QMessageBox::StandardButton result;

	sprintf(str, "Export %d visible layers to an images data file?", pointLayers.size());

	result = QMessageBox::question(this, "Confirm", str, QMessageBox::Yes|QMessageBox::No, QMessageBox::No);
	if (result != QMessageBox::Yes)
		return;

	QString fname = QFileDialog::getSaveFileName(this, "Export images data", "", "Images data files (*.imgdata)");

	if (fname.isNull())
		return;

	if (!m_pScene->exportImagesData(fname, m_pActionTriangles->isChecked(), m_pActionIntensities->isChecked(), m_pActionGroups->isChecked(), m_pActionTimeDelays->isChecked()))
	{
		QString msg = "Couldn't export to images data file: ";

		msg += m_pScene->getErrorString().c_str();

		QMessageBox::warning(this, "Error", msg);
		return;
	}
}

void GraleEditor::onAddView()
{
	EditorWidget *pEditWidget = new EditorWidget(m_pViewSplitter);	
	m_editorWidgets.append(pEditWidget);

	pEditWidget->setScene(m_pScene);
}

void GraleEditor::onDeleteView()
{
	int s;

	if ((s = m_editorWidgets.size()) < 2)
		return;

	EditorWidget *pLastView = m_editorWidgets.at(s-1);

	m_editorWidgets.removeAt(s-1);

	delete pLastView;
}

void GraleEditor::toolCheck()
{
	GraleEditor::Tool tool = getCurrentTool();

	if (tool == GraleEditor::SelectTool)
		m_pScene->enableMoveAndSelect(true);
	else if (tool == GraleEditor::GroupNumberTool || tool == GraleEditor::PointInfoTool || tool == PointMatchTool || tool == CenterTool)
		m_pScene->enableMoveAndSelect(false);
	else
		m_pScene->disableMoveAndSelect();
}

void GraleEditor::onToolChanged(QAction *pAction)
{
	toolCheck();
}

bool GraleEditor::onNew()
{
	QMessageBox::StandardButton result = QMessageBox::question(this, "Confirm", "Erase all current work?", QMessageBox::Yes|QMessageBox::No, QMessageBox::No);
	if (result != QMessageBox::Yes)
		return false;

	onNewInternal();
	return true;
}

void GraleEditor::onNewInternal()
{
	while(m_pLayersWidget->topLevelItemCount() > 0)
	{
		QTreeWidgetItem *pItem = m_pLayersWidget->takeTopLevelItem(0);
		delete pItem;

		m_pScene->deleteLayer(0);
	}

	toolCheck();
	displayActiveLayer();
}

void GraleEditor::onLoad()
{
	QString fname = QFileDialog::getOpenFileName(this, "Load", "", "GraleEditor files (*.ge)");

	if (fname.isNull())
		return;

	if(!onNew())
		return;

	onLoadInternal(fname);
}

void GraleEditor::onLoadInternal(const QString &fname)
{
	if (!m_pScene->load(fname))
	{
		QString msg = "Couldn't load layers: ";

		msg += m_pScene->getErrorString().c_str();
		QMessageBox::warning(this, "Error", msg);
		m_pScene->clear();
		return;
	}

	int numLayers = m_pScene->getNumberOfLayers();

	for (int i = 0 ; i < numLayers ; i++)
	{
		LayerItem *pLayer = m_pScene->getLayerItem(i);

		if (pLayer->getLayerType() == LayerItem::Points)
		{
			PointsLayerTreeItem *pTreeItem = new PointsLayerTreeItem(pLayer, m_pLayersWidget);

			connect(pLayer, SIGNAL(signalGroupNumberChanged()), pTreeItem, SLOT(updateProperties()));
		}
		else // image layer
		{
			ImageLayerTreeItem *pTreeItem = new ImageLayerTreeItem(pLayer, m_pLayersWidget);

			pTreeItem->initializeProperties();
			connect(pLayer, SIGNAL(signalGroupNumberChanged()), pTreeItem, SLOT(updateImageProperties()));
		}
	}

	toolCheck();
	displayActiveLayer();
}

void GraleEditor::onSave()
{
	QString fname = QFileDialog::getSaveFileName(this, "Save", "", "GraleEditor files (*.ge)");

	if (fname.isNull())
		return;

	if (!m_pScene->save(fname))
	{
		QString msg = "Couldn't save layers: ";

		msg += m_pScene->getErrorString().c_str();
		QMessageBox::warning(this, "Error", msg);
		return;
	}
}

void GraleEditor::onCopy()
{
	int idx = m_pScene->getCurrentLayer();
	LayerItem *pLayer = m_pScene->getLayerItem(idx);

	if (pLayer == 0)
		return;

//	if (pLayer->getLayerType() != LayerItem::Points)
//	{
//		QMessageBox::warning(this, "Error", "Active layer is not a points layer");
//		return;
//	}

	QList<PointItem *> points = pLayer->getSelectedPoints();
	
	m_copyPoints.clear();
	for (int i = 0 ; i < points.size() ; i++)
		m_copyPoints.append(points[i]->pos());
}

void GraleEditor::onPaste()
{
	int idx = m_pScene->getCurrentLayer();
	LayerItem *pLayer = m_pScene->getLayerItem(idx);

	if (pLayer == 0)
		return;

	if (pLayer->getLayerType() != LayerItem::Points)
	{
		QMessageBox::warning(this, "Error", "Active layer is not a points layer");
		return;
	}

	QList<PointItem *> newPoints;
	for (int i = 0 ; i < m_copyPoints.size() ; i++)
	{
		PointItem *pPoint = new PointItem(pLayer);

		pPoint->setScaleFactor(pLayer->getPointScaleFactor());
		pPoint->setPos(m_copyPoints.at(i));

		newPoints.append(pPoint);
	}
	toolCheck();

	if (getCurrentTool() == SelectTool)
	{
		for (int i = 0 ; i < newPoints.size() ; i++)
		{
			PointItem *pPoint = newPoints.at(i);
			
			pPoint->setSelected(true);
		}
	}
}

void GraleEditor::closeEvent(QCloseEvent *event)
{
	QMessageBox::StandardButton result;

	result = QMessageBox::question(this, "Confirm", "Are you sure you want to quit?", QMessageBox::Yes|QMessageBox::No, QMessageBox::No);

	if (result == QMessageBox::Yes)
	{
		event->accept();

		QSettings settings;
		settings.setValue(kWinGeom, saveGeometry());
		settings.setValue(kWinState, saveState());

		settings.setValue(kExportBitmapWidth, m_lastExportBitmapWidth);
		settings.setValue(kDefaultPointSize, m_pScene->getDefaultPointScaleFactor());
		settings.setValue(kDefaultMatchPointSize, m_pScene->getDefaultMatchPointScaleFactor());
		settings.setValue(kAxesShown, m_pScene->isCoordinateItemShown());
	}
	else
		event->ignore();
}

void GraleEditor::onExportBitmap()
{
	onExportBitmapViewOrArea(true);
}

bool GraleEditor::getVisibleImagesArea(QRectF &r)
{
	int numLayers = m_pScene->getNumberOfLayers();
	QList<LayerItem *> pointLayers;

	for (int i = 0 ; i < numLayers ; i++)
	{
		LayerItem *pLayer = m_pScene->getLayerItem(i);

		if (pLayer->isVisible() && pLayer->getLayerType() == LayerItem::Points && pLayer->getAllPoints().size() > 0)
			pointLayers.append(pLayer);
	}

	if (pointLayers.size() == 0)
	{
		QMessageBox::warning(this, "Error", "No visible point layers to base area on");
		return false;
	}

	double minX = numeric_limits<double>::infinity();
	double maxX = - numeric_limits<double>::infinity();
	double minY = numeric_limits<double>::infinity();
	double maxY = - numeric_limits<double>::infinity();

	for (int i = 0 ; i < pointLayers.size() ; i++)
	{
		LayerItem *pLayer = pointLayers.at(i);
		QList<PointItem *> pointList;
		
		pointList = pLayer->getAllPoints();
		for (int p = 0 ; p < pointList.size() ; p++)
		{
			PointItem *pPoint = pointList.at(p);
			QPointF pos = pPoint->pos();
			
			double x = pos.x();
			double y = pos.y();

			if (x > maxX) maxX = x;
			if (x < minX) minX = x;
			if (y > maxY) maxY = y;
			if (y < minY) minY = y;
		}
	}
	
	r = QRectF(minX, minY, maxX-minX, maxY-minY);

	return true;
}

void GraleEditor::onAreaExportBitmap()
{
	onExportBitmapViewOrArea(false);
}

void GraleEditor::onExportBitmapViewOrArea(bool view)
{
	QString fname = QFileDialog::getSaveFileName(this, "Save bitmap", "", "Image files (*.png *.jpg)");

	if (fname.isNull())
		return;

	bool ok;

	//double scaleFactor = QInputDialog::getDouble(this, "Enter size scale factor", "Factor", 1.0, 0.01, 100.0, 5, &ok);
	int desiredWidth = QInputDialog::getInt(this, "Width?", "Number of pixels wide:", m_lastExportBitmapWidth, 2, 16384, 1, &ok); 

	if (!ok)
		return;

	m_lastExportBitmapWidth = desiredWidth;

	QRectF viewport;

	if (view)
	{
		QGraphicsView *pFirstView = m_pScene->views().at(0);
		QRect v = pFirstView->viewport()->rect();
		QPointF p1 = pFirstView->mapToScene(v.topLeft());
		QPointF p2 = pFirstView->mapToScene(v.bottomRight());

		double x = std::min(p1.x(), p2.x());
		double y = std::min(p1.y(), p2.y());
		double w = std::abs(p1.x()-p2.x());
		double h = std::abs(p1.y()-p2.y());
		viewport = QRectF(x, y, w, h);
	}
	else
	{
		if (!getVisibleImagesArea(viewport))
			return;
	}

	//cout << viewport.x() << " " << viewport.y() << " " << viewport.width() << " " << viewport.height() << endl;

	int width = desiredWidth;
	double scaleFactor = (double)width/viewport.width();
	int height = (int)(viewport.height()*scaleFactor + 0.5);

	QPixmap pixmap(width, height);

	pixmap.fill(Qt::black);
	QPainter painter(&pixmap);

	m_pScene->setCaching(false);

	painter.setRenderHint(QPainter::Antialiasing);
	m_pScene->render(&painter, QRectF(0, 0, width, height), viewport);
	painter.end();

	m_pScene->setCaching(true);

	QImage img = pixmap.toImage().mirrored(true, true);
	img.save(fname);
}

void GraleEditor::onDefaultPointSizeChanged()
{
	QString txt = m_pDefaultPointSizeEdit->text();
	double val = txt.toDouble();

	m_pScene->setDefaultPointScaleFactor(val);

	txt.setNum(val, 'f', 10);

	m_pDefaultPointSizeEdit->setText(txt);
}

void GraleEditor::onDefaultMatchPointSizeChanged()
{
	QString txt = m_pDefaultMatchPointSizeEdit->text();
	double val = txt.toDouble();

	m_pScene->setDefaultMatchPointScaleFactor(val);

	txt.setNum(val, 'f', 10);

	m_pDefaultMatchPointSizeEdit->setText(txt);
}

void GraleEditor::onSetPointSize()
{
	bool ok;

	double s = QInputDialog::getDouble(this, "Point size", "Size", m_lastSetPointSize, 0.00001, 100.0, 10, &ok);

	if (!ok)
		return;

	m_lastSetPointSize = s;
	m_pScene->setAllPointSizes(s);

	for (int i = 0 ; i < m_pLayersWidget->topLevelItemCount() ; i++)
	{
		LayerTreeItem *pLayerTreeItem = (LayerTreeItem *)m_pLayersWidget->topLevelItem(i);
			
		if (pLayerTreeItem->getLayer()->getLayerType() == LayerItem::Points)
		{
			PointsLayerTreeItem *pPointsLayer = (PointsLayerTreeItem *)pLayerTreeItem;

			pPointsLayer->updateProperties();
		}
	}
}

void GraleEditor::onSetMatchPointSize()
{
	bool ok;
	double s = QInputDialog::getDouble(this, "Match point size", "Size", m_lastSetMatchPointSize, 0.00001, 100.0, 10, &ok);

	if (!ok)
		return;

	m_lastSetMatchPointSize = s;
	m_pScene->setAllMatchPointSizes(s);

	for (int i = 0 ; i < m_pLayersWidget->topLevelItemCount() ; i++)
	{
		LayerTreeItem *pLayerTreeItem = (LayerTreeItem *)m_pLayersWidget->topLevelItem(i);
			
		if (pLayerTreeItem->getLayer()->getLayerType() == LayerItem::Image)
		{
			ImageLayerTreeItem *pImageLayer = (ImageLayerTreeItem *)pLayerTreeItem;

			pImageLayer->updateImageProperties();
		}
	}
}

void GraleEditor::onOrientationChanged(QAction *pAction)
{
	bool left ;

	if (pAction == m_pXLeftAction)
		left = true;
	else
		left = false;

	size_t s = m_editorWidgets.size();

	for (size_t i = 0 ; i < s ; i++)
	{
		EditorWidget *pView = m_editorWidgets.at(i);

		pView->setXAxisOrientation(left);
	}	
}



