#include "hullprocessor.h"
#include <grale/polygon2d.h>
#include <QList>
#include <iostream>
#include <cmath>
#include <vector>
#include <list>
#include <assert.h>

//#include <fstream>

using namespace std;
using namespace grale;

HullProcessor::HullProcessor()
{ 
	m_scale = 1.0;
}

HullProcessor::~HullProcessor()
{ 
}

double HullProcessor::getDistance(QPointF p1, QPointF p2, QPointF p0)
{
	double x1 = p1.x(), y1 = p1.y();
	double x2 = p2.x(), y2 = p2.y();
	double x0 = p0.x(), y0 = p0.y();
	double dx = x2-x1;
	double dy = y2-y1;
	double denom = sqrt(dx*dx+dy*dy);

	return std::abs(dy*x0-dx*y0+x2*y1-y2*x1)/denom;
}

bool HullProcessor::createImgDataPointsFromHull(const QVector<QPointF> &pts)
{
	/*
	ofstream debugFile;

	debugFile.open("lasthull.txt");
	debugFile << "double hullPoints[][2] = {" << endl;
	for (const QPointF &p : pts)
	{
		debugFile << "    { " << p.x() << ", " << p.y() << " }," << endl; 
	}
	debugFile << "};" << endl;
	debugFile.close();
	*/

	if (pts.size() < 3)
	{
		setErrorString("At least three points are needed");
		return false;
	}

	double scale = getScale(pts);

	// Calculate relative distance of each part
	const int numPts = pts.size();

	// Start with a few points on the hull
	int idxOneThird = numPts/3;
	int idxTwoThirds = numPts*2/3;
	if (!(0 < idxOneThird && idxOneThird < idxTwoThirds && idxTwoThirds < numPts))
	{
		setErrorString("Couldn't find three separate points to start from");
		return false;
	}

	QVector<int> imgPoints = { 0, idxOneThird, idxTwoThirds, numPts };

	// Now refine
	const int maxImgPoints = 50; // TODO: make this configurable
	bool approximationDone = false;
	const double maxToleratedDistance = scale/50.0; // TODO: make this configurable

	while (imgPoints.size() < maxImgPoints && !approximationDone)
	{
		QVector<int> newImgPoints;
		bool allDistancesOk = true;

		newImgPoints.push_back(0);
		//cout << endl;
		for (int i = 0 ; i < imgPoints.size()-1 && newImgPoints.size() < maxImgPoints ; i++)
		{
			const int idx0 = imgPoints[i];
			const int idx1 = imgPoints[i+1];
			const int numInBetween = idx1-idx0-1;

			//cout << "idx0 = " << idx0 << " idx1 = " << idx1 << endl;
			assert(idx1 > idx0);

			if (numInBetween > 0)
			{
				assert(idx0 >= 0 && idx0 < numPts && idx1 >= 0 && idx1 <= numPts);

				QPointF p0 = pts[idx0];
				QPointF p1 = pts[idx1%numPts];

				double maxDist = 0;
				int maxDistIdx = -1;
				for (int j = 0 ; j < numInBetween ; j++)
				{
					int idx = idx0+1+j;
					assert(idx >= 0 && idx < numPts);
					QPointF testPoint = pts[idx];

					double dist = getDistance(p0, p1, testPoint);
					if (dist > maxDist)
					{
						maxDist = dist;
						maxDistIdx = idx;
					}
				}

				if (maxDist > maxToleratedDistance)
				{
					newImgPoints.push_back(maxDistIdx);
					allDistancesOk = false;
				}
			}

			newImgPoints.push_back(idx1);
		}

		imgPoints = newImgPoints;

		if (allDistancesOk)
			approximationDone = true;
	}
	
	// Store these points
	m_imgDataPoints.clear();
	for (auto idx : imgPoints)
	{
		if (idx < numPts) // The last one is just for convenience during calculation, don't store it
		{
			assert(idx >= 0 && idx < numPts);
			m_imgDataPoints.push_back(pts[idx]);
		}
	} 

	// Create triangulation, remove triangles which are not inside

	if (!createTriangulation())
		return false; // error already set

	m_scale = scale;
	return true;
}

bool HullProcessor::createTriangulation()
{
	if (m_imgDataPoints.size() < 3)
	{
		setErrorString("Not enough points to create a triangulation");
		return false;
	}

	vector<Vector2D<double>> polygonPoints;
	for (auto p : m_imgDataPoints)
		polygonPoints.push_back(Vector2D<double>(p.x(), p.y()));

	// We'll use this polygon to remove triangles that are on the outside
	Polygon2D<double> polygon;
	polygon.init(&polygonPoints[0], polygonPoints.size(), false);

	Triangulation triang;

	if (!triang.create(polygonPoints))
	{
		setErrorString("Can't create triangulation: " + triang.getErrorString());
		return false;
	}

	list<TriangleIndices> triangles;

	triang.getTriangles(triangles);

	// Only keep the triangles of which the centroid is inside the polygon
	m_triangulation.clear();
	for (const TriangleIndices &t : triangles)
	{
		Vector2D<double> p0 = polygonPoints[t.getIndex(0)];
		Vector2D<double> p1 = polygonPoints[t.getIndex(1)];
		Vector2D<double> p2 = polygonPoints[t.getIndex(2)];
		double xc = (p0.getX() + p1.getX() + p2.getX())/3.0;
		double yc = (p0.getY() + p1.getY() + p2.getY())/3.0;

		if (polygon.isInside(Vector2D<double>(xc, yc)))
			m_triangulation.push_back(t);
	}
	return true;
}

double HullProcessor::getScale(const QVector<QPointF> &pts)
{
	QPointF firstPoint = pts[0];
	double minX = firstPoint.x(), maxX = firstPoint.x();
	double minY = firstPoint.y(), maxY = firstPoint.y();

	for (const QPointF &p : pts)
	{
		if (p.x() < minX) minX = p.x();
		if (p.x() > maxX) maxX = p.x();
		if (p.y() < minY) minY = p.y();
		if (p.y() > maxY) maxY = p.y();
	}

	double dx = maxX-minX;
	double dy = maxY-minY;

	return sqrt(dx*dx+dy*dy);
}

