/*

  This file is a part of GRALEEDITOR, a program to help create input data
  for gravitational lens inversion using GRALE.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "layeritem.h"
#include "pointitem.h"
#include "triangleitem.h"
#include "pointmatchitem.h"
#include "hullitem.h"
#include "hullprocessor.h"
#include "graleeditor.h"
#include <QPainter>
#include <QMessageBox>
#include <QStyleOptionGraphicsItem>

#include <iostream>

using namespace std;
using namespace grale;

LayerItem::LayerItem(LayerType t) : m_layerType(t)
{
	enable();

	m_pointScaleFactor = 0.01;
	m_nextGroupNumber = 0;
	m_hullDrawing = false;
	
	if (t == Points)
		m_pHullItem = new HullItem(this);
	else
		m_pHullItem = nullptr;
}

LayerItem::~LayerItem()
{
}

QRectF LayerItem::boundingRect() const
{
	return childrenBoundingRect();
}

void LayerItem::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget)
{
	/*
	painter->setPen(Qt::NoPen);
	painter->setBrush(Qt::darkGray);
	painter->drawRect(childrenBoundingRect());
	*/
}

void LayerItem::addPoint(QPointF pos)
{
	if (m_layerType == Points)
	{
		PointItem *pPoint = new PointItem(this);
		pPoint->setPos(pos);
		pPoint->setScaleFactor(m_pointScaleFactor);
	}
}

void LayerItem::setPointScaleFactor(double factor)
{
	m_pointScaleFactor = factor;

	QList<QGraphicsItem *> children = childItems();
	QList<PointItem *> points;

	for (int i = 0 ; i < children.size() ; i++)
	{
		QGraphicsItem *pChild = children.at(i);
		PointItem *pPoint = dynamic_cast<PointItem *>(pChild);

		if (pPoint != 0)
			pPoint->setScaleFactor(m_pointScaleFactor);

		QList<QGraphicsItem *> grandChildren = pChild->childItems();

		for (int j = 0 ; j < grandChildren.size() ; j++)
		{
			QGraphicsItem *pGrandChild = grandChildren.at(j);
			PointMatchItem *pPoint2 = dynamic_cast<PointMatchItem *>(pGrandChild);

			if (pPoint2 != 0)
				pPoint2->setScaleFactor(m_pointScaleFactor);
		}
	}
}

void LayerItem::disable()
{
	QList<QGraphicsItem *> children = childItems();

	for (int i = 0 ; i < children.size() ; i++)
		children.at(i)->setAcceptedMouseButtons(Qt::NoButton);

	setAcceptedMouseButtons(Qt::NoButton);
	m_enabled = false;
}

void LayerItem::enable()
{
	QList<QGraphicsItem *> children = childItems();

	for (int i = 0 ; i < children.size() ; i++)
		children.at(i)->setAcceptedMouseButtons(Qt::LeftButton|Qt::RightButton|Qt::MidButton);

	setAcceptedMouseButtons(Qt::LeftButton|Qt::RightButton|Qt::MidButton);
	m_enabled = true;
}

QList<PointItem *> LayerItem::getSelectedPoints()
{
	QList<QGraphicsItem *> children = childItems();
	QList<PointItem *> points;

	for (int i = 0 ; i < children.size() ; i++)
	{
		QGraphicsItem *pChild = children.at(i);

		if (pChild->isSelected())
		{
			PointItem *pPoint = dynamic_cast<PointItem *>(pChild);

			if (pPoint != 0)
				points.append(pPoint);
		}
	}

	return points;
}

QList<TriangleItem *> LayerItem::getSelectedTriangles()
{
	QList<QGraphicsItem *> children = childItems();
	QList<TriangleItem *> triangles;

	for (int i = 0 ; i < children.size() ; i++)
	{
		QGraphicsItem *pChild = children.at(i);

		if (pChild->isSelected())
		{
			TriangleItem *pTriangle = dynamic_cast<TriangleItem *>(pChild);

			if (pTriangle != 0)
				triangles.append(pTriangle);
		}
	}

	return triangles;
}

QList<PointItem *> LayerItem::getAllPoints()
{
	QList<QGraphicsItem *> children = childItems();
	QList<PointItem *> points;

	for (int i = 0 ; i < children.size() ; i++)
	{
		QGraphicsItem *pChild = children.at(i);
		PointItem *pPoint = dynamic_cast<PointItem *>(pChild);

		if (pPoint != 0)
			points.append(pPoint);
	}

	return points;
}

QList<TriangleItem *> LayerItem::getAllTriangles()
{
	QList<QGraphicsItem *> children = childItems();
	QList<TriangleItem *> triangles;

	for (int i = 0 ; i < children.size() ; i++)
	{
		QGraphicsItem *pChild = children.at(i);
		TriangleItem *pTriangle = dynamic_cast<TriangleItem *>(pChild);

		if (pTriangle != 0)
			triangles.append(pTriangle);
	}

	return triangles;
}

void LayerItem::deleteTriangles()
{
	QList<QGraphicsItem *> children = childItems();

	for (int i = 0 ; i < children.size() ; i++)
	{
		QGraphicsItem *pChild = children.at(i);
		TriangleItem *pTriangle = dynamic_cast<TriangleItem *>(pChild);

		if (pTriangle != 0)
			delete pTriangle;
	}
}

void LayerItem::setGroupNumber(int n)
{
	m_nextGroupNumber = n;

	emit signalGroupNumberChanged();
}

QList<PointMatchItem *> LayerItem::getSelectedMatchPoints()
{
	QList<QGraphicsItem *> children = childItems();
	QList<PointMatchItem *> points;

	for (int i = 0 ; i < children.size() ; i++)
	{
		QGraphicsItem *pChild = children.at(i); // is a FITS item or image item
		QList<QGraphicsItem *> grandChildren = pChild->childItems();

		for (int j = 0 ; j < grandChildren.size() ; j++)
		{
			QGraphicsItem *pGrandChild = grandChildren.at(j);

			if (pGrandChild->isSelected())
			{
				PointMatchItem *pPoint = dynamic_cast<PointMatchItem *>(pGrandChild);
	
				if (pPoint != 0)
					points.append(pPoint);
			}
		}
	}
	return points;
}

QList<PointMatchItem *> LayerItem::getAllMatchPoints()
{
	QList<QGraphicsItem *> children = childItems();
	QList<PointMatchItem *> points;

	for (int i = 0 ; i < children.size() ; i++)
	{
		QGraphicsItem *pChild = children.at(i); // is a FITS item or image item
		QList<QGraphicsItem *> grandChildren = pChild->childItems();

		for (int j = 0 ; j < grandChildren.size() ; j++)
		{
			QGraphicsItem *pGrandChild = grandChildren.at(j);
			PointMatchItem *pPoint = dynamic_cast<PointMatchItem *>(pGrandChild);

			if (pPoint != 0)
				points.append(pPoint);
		}
	}
	return points;
}

void LayerItem::startHull(QPointF p)
{
	if (m_pHullItem)
	{
		m_pHullItem->startHull(p);
		m_hullDrawing = true;
	}
}

void LayerItem::endHull(bool cancel)
{
	if (m_pHullItem)
	{
		if (!cancel)
		{
			auto hullPoints = m_pHullItem->getHullPoints();

			HullProcessor hp;

			if (!hp.createImgDataPointsFromHull(hullPoints))
			{
				//QWidget *pParent = GraleEditor::getInstance();
				//QMessageBox::warning(pParent, "Error", QString(hp.getErrorString().c_str()));
			}
			else
			{
				auto pts = hp.getImageDataPoints();
				auto triangles = hp.getTriangulation();
				double scale = hp.getScale();
				double pointScale = scale/200.0;
				QVector<PointItem *> pointItems;

				for (const QPointF &p : pts)
				{
					PointItem *pPoint = new PointItem(this);
					pPoint->setPos(p);
					pPoint->setScaleFactor(pointScale);

					pointItems.push_back(pPoint);
				}

				for (const TriangleIndices &t : triangles)
				{
					auto p0 = pointItems[t.getIndex(0)];
					auto p1 = pointItems[t.getIndex(1)];
					auto p2 = pointItems[t.getIndex(2)];

					TriangleItem *pTreeItem = new TriangleItem(p0, p1, p2, this);	
				}
			}
		}
		m_pHullItem->endHull();
		m_hullDrawing = false;
	}
}

void LayerItem::addHullPoint(QPointF p)
{
	if (m_pHullItem && m_hullDrawing)
		m_pHullItem->addHullPoint(p);
}

