#include <QApplication>
#include <QImage>
#include <iostream>
#include <limits>
#include <cmath>
#include <algorithm>
#include "contourfinder.h"

using namespace std;
using namespace grale;

vector<double> blur(const vector<double> &values, int w, int h, int kernelWidth)
{
	const int kernelSize = 2*kernelWidth+1;
	const int kernelCenter = kernelWidth;
	vector<double> kernel(kernelSize*kernelSize);

	double norm = 0;
	for (int y = 0 ; y < kernelSize ; y++)
	{
		double dy = y-kernelCenter;
		for (int x = 0 ; x < kernelSize ; x++)
		{
			double dx = x-kernelCenter;
			double val = exp(-(dx*dx+dy*dy)/(2.0*(kernelWidth*3)));
			norm += val;
			kernel[x+y*kernelSize] = val;
		}
	}

	for (auto &x : kernel)
		x /= norm;

	// TODO: check negative!
	const int resultW = w - kernelWidth*2;
	const int resultH = h - kernelWidth*2;
	vector<double> convolution(resultW*resultH);
	for (int y = 0 ; y < resultH ; y++)
	{
		for (int x = 0 ; x < resultW ; x++)
		{
			double sum = 0;
			for (int ky = 0 ; ky < kernelSize ; ky++)
			{
				for (int kx = 0 ; kx < kernelSize ; kx++)
				{
					sum += kernel[kx+ky*kernelSize]*values[(x+kx)+(y+ky)*w];
				}
			}
			convolution[x+y*resultW] = sum;
		}
	}

	return convolution;
}

#if 0
vector<vector<Vector2Dd>> closeContours(const vector<vector<Vector2Dd>> &contours)
{
	vector<vector<Vector2Dd>> inputContours = contours;
	vector<vector<Vector2Dd>> outputContours;

	double maxDist = 5; // TODO!
	bool done = false;
	while (!done)
	{
		bool changedSomething = false;

		for (int i = 0 ; !changedSomething && i < inputContours.size() ; i++)
		{
			if (inputContours[i].size() < 2)
				continue;

			Vector2Dd begin = inputContours[i][0];
			Vector2Dd end = inputContours[i][inputContours[i].size()-1];

			for (int j = 0 ; !changedSomething && j < inputContours.size() ; j++)
			{
				if (j == i || inputContours[j].size() < 2)
					continue;

				Vector2Dd d1 = begin;
				d1 -= inputContours[j][0];

				Vector2Dd d2 = end;
				d2 -= inputContours[j][0];
				
				if (d1.getLength() < maxDist || d2.getLength() < maxDist)
				{
					cout << "# d1 = " << d1.getLength() << " d2 = " << d2.getLength() << endl;
					cout << "# appending contours " << i << " and " << j<< endl;
					vector<Vector2Dd> startContour = inputContours[i];
					vector<Vector2Dd> appendContour = inputContours[j];

					if (d1.getLength() < maxDist) // need to reverse startContour first
						reverse(startContour.begin(), startContour.end());
						
					startContour.insert(startContour.end(), appendContour.begin(), appendContour.end());

					outputContours.push_back(startContour);
					for (int k = 0 ; k < inputContours.size() ; k++)
					{
						if (k == i || k == j)
							continue;
						outputContours.push_back(inputContours[k]);
					}
					changedSomething = true;
				}
			}
		}

		if (!changedSomething)
			done = true;
		else
		{
			inputContours.swap(outputContours);
			outputContours.clear();
		}
	}

	return inputContours;
}
#endif

int main(int argc, char *argv[])
{
	QApplication app(argc, argv);
	QImage img;

	if (!img.load("testimage.png"))
	{
		cerr << "Unable to load image" << endl;
		return -1;
	}

	int w = img.width();
	int h = img.height();
	vector<double> values(w*h);

	for (int y = 0 ; y < h ; y++)
	{
		for (int x = 0 ; x < w ; x++)
			values[x+y*w] = qGray(img.pixel(x, y));
	}

	int kWidth = 2;
	values = blur(values, w, h, kWidth);
	w -= kWidth*2;
	h -= kWidth*2;

	ContourFinder cf(values, Vector2Dd(0.5, 0.5+(h-1)), Vector2Dd(0.5+(w-1), 0.5), w, h);
	Vector2Dd center(w/2.0, h/2.0);
	vector<vector<Vector2Dd>> contours(256);
	vector<double> averageDistances(contours.size());

	for (int i = 0 ; i < contours.size() ; i++)
	{
		if (!cf.findContours(0.5 + i))
		{
			cerr << "Error finding contours: " << cf.getErrorString() << endl;
			return -1;
		}

		//vector<vector<Vector2Dd>> ctrList = closeContours(cf.getContours());
		const vector<vector<Vector2Dd>> &ctrList = cf.getContours();
		double bestAvgDist = numeric_limits<double>::max();
		vector<Vector2Dd> closestContour;

		for (const vector<Vector2Dd> &ctr : ctrList)
		{
			double avgDist = 0;
			for (Vector2Dd pos : ctr)
			{
				Vector2Dd diff = pos;
				diff -= center;

				avgDist += diff.getLength();
				//cout << pos.getX() << " " << pos.getY() << endl;
			}
			avgDist /= ctr.size();

			//cout << endl;

			if (avgDist < bestAvgDist)
			{
				bestAvgDist = avgDist;
				closestContour = ctr;
			}
		}

		contours[i] = closestContour;
		averageDistances[i] = bestAvgDist;
		//cout << endl;
	}

	for (int i = 0 ; i < contours.size() ; i++)
	{
		if (averageDistances[i] < numeric_limits<double>::max())
			cout << averageDistances[i] << " " << (0.5+i) << endl;
	}
	cout << endl << endl;

	for (int i = 0 ; i < contours.size() ; i++)
	{
		const vector<Vector2Dd> &ctr = contours[i];
		
		for (Vector2Dd pos : ctr)
			cout << pos.getX() << " " << pos.getY() << endl;
		
		cout << endl;
		cout << endl;
	}

	return 0;
}

