/*

  This file is a part of GRALEEDITOR, a program to help create input data
  for gravitational lens inversion using GRALE.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "pointitem.h"
#include "triangleitem.h"
#include "graleeditor.h"
#include "layeritem.h"
#include "pointinfodialog.h"
#include <QPainter>
#include <QStyleOptionGraphicsItem>
#include <QGraphicsTextItem>
#include <QColor>


#include <iostream>

using namespace std;

PointItem::PointItem(QGraphicsItem * pParent) : QGraphicsItem(pParent)
{
	// setFlag(ItemIsMovable);
	// setFlag(ItemIsSelectable);

	m_pGroupTextItem = 0;
	m_pTimeTextItem = 0;

	m_intensity = 0;
	m_group = -1;
	m_timeDelay = 0;
}

PointItem::~PointItem()
{
}

void PointItem::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget)
{
	painter->setPen(Qt::NoPen);
	painter->setBrush(Qt::darkGray);
	painter->drawEllipse(-7, -7, 20, 20);

	QRadialGradient gradient(-3, -3, 10);
	if (option->state & QStyle::State_Sunken) 
	{
		gradient.setCenter(3, 3);
		gradient.setFocalPoint(3, 3);
		gradient.setColorAt(1, QColor(Qt::yellow).light(120));
		gradient.setColorAt(0, QColor(Qt::darkYellow).light(120));
	} 
	else 
	{
		if (isSelected())
		{
			gradient.setColorAt(0, Qt::blue);
			gradient.setColorAt(1, Qt::darkBlue);
		}
		else
		{
			gradient.setColorAt(0, Qt::yellow);
			gradient.setColorAt(1, Qt::darkYellow);
		}
	}
	painter->setBrush(gradient);
	painter->setPen(QPen(Qt::black, 0));
	painter->drawEllipse(-10, -10, 20, 20);

}

QRectF PointItem::boundingRect() const
{
	qreal adjust = 2;
	QRectF rect = QRectF(-10 - adjust, -10 - adjust, 23 + adjust, 23 + adjust);
	return rect.united(childrenBoundingRect());
}

QPainterPath PointItem::shape() const
{
	QPainterPath path;
	
	path.addEllipse(-10, -10, 20, 20);
	return path;
}


void PointItem::mousePressEvent(QGraphicsSceneMouseEvent *event)
{
	for (int i = 0 ; i < m_triangles.size() ; i++)
	{
		TriangleItem *pTriangle = m_triangles.at(i);
		
		pTriangle->pointChanged();
	}
	update();
	QGraphicsItem::mousePressEvent(event);
}

void PointItem::mouseMoveEvent(QGraphicsSceneMouseEvent *event)
{
	for (int i = 0 ; i < m_triangles.size() ; i++)
	{
		TriangleItem *pTriangle = m_triangles.at(i);
		
		pTriangle->pointChanged();
	}
	QGraphicsItem::mouseMoveEvent(event);
}

void PointItem::mouseReleaseEvent(QGraphicsSceneMouseEvent *event)
{
	GraleEditor::Tool tool = GraleEditor::getInstance()->getCurrentTool();

	for (int i = 0 ; i < m_triangles.size() ; i++)
	{
		TriangleItem *pTriangle = m_triangles.at(i);
		
		pTriangle->pointChanged();
	}

	if (tool == GraleEditor::GroupNumberTool)
	{
		LayerItem *pLayer = (LayerItem *)parentItem();
		int number = pLayer->getNextGroupNumber();

		setGroupNumber(number);

		if (number >= 0)
			pLayer->setGroupNumber(number+1);
	}
	else if (tool == GraleEditor::PointInfoTool)
	{
		PointInfoDialog pointInfo(getIntensity(), hasTimeDelay(), getTimeDelay(), GraleEditor::getInstance());

		if (pointInfo.exec() == QDialog::Accepted)
		{
			if (pointInfo.hasTimeDelay())
				setTimeDelay(pointInfo.getTimeDelay());
			else
				clearTimeDelay();

			setIntensity(pointInfo.getIntensity());
		}
	}

	update();
	QGraphicsItem::mouseReleaseEvent(event);
}

QVariant PointItem::itemChange(GraphicsItemChange change, const QVariant & value)
{
	for (int i = 0 ; i < m_triangles.size() ; i++)
	{
		TriangleItem *pTriangle = m_triangles.at(i);
		
		pTriangle->pointChanged();
	}
	return QGraphicsItem::itemChange(change, value);
}

void PointItem::setScaleFactor(double scale)
{
	setTransform(QTransform(-scale, 0, 0, -scale, 0, 0), false);
}

void PointItem::setGroupNumber(int id)
{
	m_group = id;

	if (id >= 0)
	{
		QString idStr = QString::number(id, 10);

		if (m_pGroupTextItem == 0)
		{
			m_pGroupTextItem = new QGraphicsTextItem(this);
			m_pGroupTextItem->setDefaultTextColor(QColor(0, 0, 255));
			m_pGroupTextItem->setZValue(0.1);
		}
		m_pGroupTextItem->setPlainText(idStr);

		QRectF r = m_pGroupTextItem->boundingRect();

		m_pGroupTextItem->setPos(-r.width()/2.0, -r.height()/2.0);
	}
	else
	{
		if (m_pGroupTextItem != 0)
		{
			delete m_pGroupTextItem;
			m_pGroupTextItem = 0;
		}
	}

	prepareGeometryChange();
}

void PointItem::setTimeDelay(double td)
{
	m_timeDelay = td;

	QString idStr = QString::number(td);

	idStr += QString(" days");

	if (m_pTimeTextItem == 0)
	{
		m_pTimeTextItem = new QGraphicsTextItem(this);
		m_pTimeTextItem->setDefaultTextColor(QColor(0,255,0));
		m_pTimeTextItem->setZValue(0.2);
		setZValue(0.01);
	}

	m_pTimeTextItem->setPlainText(idStr);

	QRectF r = m_pTimeTextItem->boundingRect();

	m_pTimeTextItem->setPos(-r.width()/2.0, 11);
}

void PointItem::clearTimeDelay()
{
	if (m_pTimeTextItem == 0)
		return;

	delete m_pTimeTextItem;
	m_pTimeTextItem = 0;

	setZValue(0);
}

