#include "pointsmenu.h"
#include "layeritem.h"
#include "graleeditor.h"
#include "editorwidget.h"
#include "pointitem.h"
#include "triangleitem.h"
#include "hullprocessor.h"
#include <grale/polygon2d.h>
#include <QGraphicsItem>
#include <QAction>
#include <QMessageBox>
#include <iostream>
#include <cmath>
#include <assert.h>

using namespace std;
using namespace grale;

PointsMenu::PointsMenu(LayerItem *pLayerItem) : QMenu(nullptr), m_pLayerItem(pLayerItem)
{
	QAction *pCenter = new QAction("Center on points", this);
	addAction(pCenter);

	QAction *pExpand = new QAction("Create enlarged image", this);
	addAction(pExpand);

	QObject::connect(pCenter, &QAction::triggered, this, &PointsMenu::onCenterOnPoints);
	QObject::connect(pExpand, &QAction::triggered, this, &PointsMenu::onExpandPoints);
}

PointsMenu::~PointsMenu()
{
}

void PointsMenu::onCenterOnPoints()
{
	assert(m_pLayerItem);
	
	auto points = m_pLayerItem->getAllPoints();
	if (points.empty())
	{
		QMessageBox::warning(GraleEditor::getInstance(), "Error", "No point available to center on");
		return;
	}

	double xSum = 0.0;
	double ySum = 0.0;
	for (auto p : points)
	{
		assert(p);

		QPointF pos = p->pos();
		xSum += pos.x();
		ySum += pos.y();
	}

	xSum /= points.size();
	ySum /= points.size();

	for (auto w : GraleEditor::getInstance()->getEditorWidgets())
		w->centerOn(xSum, ySum);
}

void PointsMenu::onExpandPoints()
{
	// TODO: For now we'll use a convex hull
	
	double addedRadius = 1.0; // TODO: make this configurable

	QList<PointItem *> allPoints = m_pLayerItem->getAllPoints();
	if (allPoints.size() < 3)
	{
		QMessageBox::warning(GraleEditor::getInstance(), "Error", "Need at least three points");
		return;
	}

	vector<Vector2D<double>> points;
	for (auto p : allPoints)
	{	
		QGraphicsItem *pItem = p;
		QPointF pos = pItem->pos();
		points.push_back(Vector2D<double>(pos.x(), pos.y()));
	}

	// First canculate a convex hull

	Polygon2D<double> convexHull;
	convexHull.init(&points[0], points.size(), true);

	// Then, for each point on the convex hull, add a number of points on a circle
	// (with this point as center)
	vector<Vector2D<double>> newPoints;

	{
		const Vector2D<double> *pHullPoints = convexHull.getPoints();
		const int numHullPoints = convexHull.getNumberOfPoints();

		for (int i = 0 ; i < numHullPoints ; i++)
		{
			Vector2D<double> center = pHullPoints[i];
			const int numCirclePoints = 20;
			const double angleStep = 2.0*M_PI/numCirclePoints;

			for (int j = 0 ; j < numCirclePoints ; j++)
			{
				double angle = j*angleStep;
				double x = addedRadius*cos(angle) + center.getX();
				double y = addedRadius*sin(angle) + center.getY();

				newPoints.push_back(Vector2D<double>(x, y));
			}
		}
	}

	// Then, calculate the new convex hull of these points

	Polygon2D<double> newConvexHull;
	newConvexHull.init(&newPoints[0], newPoints.size(), true);

	// Clear everything in this layer, add these points and a triangulation

	// First delete the triangles, as they depend on other items
	QList<QGraphicsItem *> children = m_pLayerItem->childItems();
	for (int i = 0 ; i < children.size() ; i++)
	{
		QGraphicsItem *pChild = children.at(i);
	    TriangleItem *pTriangle = dynamic_cast<TriangleItem *>(pChild);
        if (pTriangle)
			delete pTriangle;
	}

	// Then clear the rest
	children = m_pLayerItem->childItems();
	for (int i = 0 ; i < children.size() ; i++)
	{
		QGraphicsItem *pChild = children.at(i);
		delete pChild;
	}

	// Add the points from the hull
	{
		const Vector2D<double> *pHullPoints = newConvexHull.getPoints();
		const int numHullPoints = newConvexHull.getNumberOfPoints();
		QVector<QPointF> qtPoints;

		for (int i = 0 ; i < numHullPoints ; i++)
			qtPoints.push_back(QPointF(pHullPoints[i].getX(), pHullPoints[i].getY()));

		HullProcessor hp;
		if (!hp.createImgDataPointsFromHull(qtPoints))
		{
			QMessageBox::warning(GraleEditor::getInstance(), "Error", QString("Unable to process new hull: ") + 
					             QString(hp.getErrorString().c_str()));
			return;
		}

		const double pointScaleFactor = hp.getScale()/200; // TODO: make this configurable

		const QVector<QPointF> &hullPoints = hp.getImageDataPoints();
		const QVector<TriangleIndices> &triang = hp.getTriangulation();
		QVector<PointItem *> pointItems;

		for (auto p : hullPoints)
		{
			PointItem *pItem = new PointItem(m_pLayerItem);
			pItem->setPos(p);
			pItem->setScaleFactor(pointScaleFactor);
			pointItems.push_back(pItem);
		}

		for (auto t : triang)
		{
			int idx0 = t.getIndex(0);
			int idx1 = t.getIndex(1);
			int idx2 = t.getIndex(2);

			TriangleItem *pTriangItem = new TriangleItem(pointItems[idx0], pointItems[idx1], pointItems[idx2], m_pLayerItem);
		}
	}
}

